package org.softevo.apibirthmark.asmadapters;

import org.objectweb.asm.Label;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;

import org.softevo.apibirthmark.constants.Constants;

import org.softevo.apibirthmark.asmadapters.proxies.GenerateProxyClassAdapter;

import org.softevo.apibirthmark.tracing.mapping.ClassMethodID;
import org.softevo.apibirthmark.tracing.mapping.NameToIDMap;

/**
 * The BytecodeTasks class incorporates several common bytecode transformations that
 * are used by different MethodAdapters.
 *
 *
 * @author schuler
 *
 */

public class BytecodeTasks implements Opcodes {


	/**
	 * Location of the tracing class at runtime
	 */
	public static final String TRACER_CLASS = "org/softevo/apibirthmark/tracing/Tracer";

	/**
	 * Inserts bytecode in given MethodVisitor to print given message.
	 *
	 * @param mv
	 * @param message
	 */
	public static void insertPrintlnlog(MethodVisitor mv, String message) {
		Label l0 = new Label();
		mv.visitLabel(l0);
		mv.visitFieldInsn(GETSTATIC, "java/lang/System", "err",
				"Ljava/io/PrintStream;");
		Label l1 = new Label();
		mv.visitJumpInsn(IFNULL, l1);
		Label l2 = new Label();
		mv.visitLabel(l2);
		mv.visitFieldInsn(GETSTATIC, "java/lang/System", "err",
				"Ljava/io/PrintStream;");
		mv.visitLdcInsn("FUNCTION CALLED \t" + message);
		mv.visitMethodInsn(INVOKEVIRTUAL, "java/io/PrintStream", "println",
				"(Ljava/lang/String;)V");
		mv.visitLabel(l1);
	}

	/**
	 * Insert the statements necessary to perform the tracing at runtime.
	 *
	 * @param mv
	 *            MethodVisitor to insert trace statements
	 * @param className
	 *            Name of the class that is instrumented
	 * @param logClassName
	 *            Name of the class that should be logged (if user class
	 *            subclasses API)
	 * @param methodName
	 *            Name of the method that is instrumented
	 * @param desc
	 *            Method descriptor of the method that is instrumented
	 * @param idMap
	 *            NameToIDMap to resolve names to IDs
	 * @param uniqueObjectPosition
	 *            position (in local variable table) of Object containing Trace
	 *            information
	 * @param isVirtual
	 *            indicates if method is virtual or static
	 *
	 */
	private static void insertTracingStatements(MethodVisitor mv,
			String className, String logClassName, String methodName,
			String desc, NameToIDMap idMap, int uniqueObjectPosition,
			boolean isVirtual) {
		insertTracingLog(mv, className, logClassName, methodName, desc, idMap,
				uniqueObjectPosition, isVirtual);
	}

	/**
	 * Insert the statements necessary to perform the tracing at runtime.
	 *
	 * @param mv
	 *            MethodVisitor to insert trace statements
	 * @param className
	 *            Name of the class that is instrumented
	 * @param methodName
	 *            Name of the method that is instrumented
	 * @param desc
	 *            Method descriptor of the method that is instrumented
	 * @param idMap
	 *            NameToIDMap to resolve names to IDs
	 * @param uniqueObjectPosition
	 *            position (in local variable table) of Object containing Trace
	 *            information
	 * @param isVirtual
	 *            indicates if method is virtual or static
	 *
	 */
	public static void insertTracingStatements(MethodVisitor mv,
			String className, String methodName, String desc,
			NameToIDMap idMap, int uniqueObjectPosition, boolean isVirtual) {
		insertTracingLog(mv, className, className, methodName, desc, idMap,
				uniqueObjectPosition, isVirtual);
	}

	/**
	 * Inserts the bytecode in given method visitor necessary to perform the
	 * tracing at runtime.
	 *
	 * @param mv
	 *            MethodVisitor to insert trace statements
	 * @param className
	 *            Name of the class that is instrumented
	 * @param logClassName
	 *            Name of the class that should be logged (if user class
	 *            subclasses API)
	 * @param methodName
	 *            Name of the method that is instrumented
	 * @param desc
	 *            Method descriptor of the method that is instrumented
	 * @param idMap
	 *            NameToIDMap to resolve names to IDs
	 * @param uniqueObjectPosition
	 *            Local variable position of the argument that carries the tracing
	 *            information
	 * @param isVirtual
	 *            indicates if method is virtual or not
	 */
	private static void insertTracingLog(MethodVisitor mv, String className,
			String logClassName, String methodName, String desc,
			NameToIDMap idMap, int uniqueObjectPosition, boolean isVirtual) {
		if (uniqueObjectPosition != -1) {
			mv.visitVarInsn(ALOAD, uniqueObjectPosition);
			mv.visitMethodInsn(INVOKEVIRTUAL, "java/lang/MyUniqueObject",
					"getClassID", "()I");
			mv.visitVarInsn(ALOAD, uniqueObjectPosition);
			mv.visitMethodInsn(INVOKEVIRTUAL, "java/lang/MyUniqueObject",
					"getMethodID", "()I");
			mv.visitVarInsn(ALOAD, uniqueObjectPosition);
			mv.visitMethodInsn(INVOKEVIRTUAL, "java/lang/MyUniqueObject",
					"getObjectID", "()I");

		} else {

			mv.visitIntInsn(SIPUSH, 1);
			mv.visitIntInsn(SIPUSH, 1);
			mv.visitIntInsn(SIPUSH, 1);
		}
		ClassMethodID id = idMap.getClassMethodID(logClassName, methodName,
				desc);
		// Maybe performance enhancement if SIPUSH and BIPUSH are used
		mv.visitLdcInsn(new Integer(id.getClassID()));
		mv.visitLdcInsn(new Integer(id.getMethodID()));
		if (isVirtual) {
			mv.visitVarInsn(ALOAD, 0);
			mv.visitFieldInsn(GETFIELD, className,
					GenerateProxyClassAdapter.OBJECTID, "I");
		} else {
			mv.visitIntInsn(SIPUSH, Constants.STATICOBJECTID);
		}
		mv.visitMethodInsn(INVOKESTATIC, TRACER_CLASS, "traceINSTR",
				"(IIIIII)V");

	}

	/**
	 *
	 * Inserts the bytecode in given method visitor necessary to perform the
	 * tracing at runtime. Then calls to the original method for this
	 * proxy method are added, and if the original method has a return value this
	 * value is also returned.
	 *
	 * @param proxyMethod
	 *            MethodVisitor to insert statements
	 * @param className
	 *            Name of the class that is instrumented
	 * @param logClassName
	 *            Name of the class that should be logged (if user class
	 *            subclasses API)
	 * @param methodName
	 *            Name of the method that is instrumented
	 * @param access
	 *            Access modifier of the method
	 * @param methodDescriptor
	 *            Method Descriptor
	 * @param idMap
	 *            NameToIDMapthat contains the ID for the method
	 */
	public static void insertLog(MethodVisitor proxyMethod, String className,
			String logClassName, String methodName, int access,
			String methodDescriptor, NameToIDMap idMap) {
		int uniqueObjectPosition = BytecodeTasks.getUniqueObjectPosition(
				access, methodDescriptor);
		int localLoadPosition = 0;
		boolean isVirtual = false;
		/* if Method is not static push this reference on the stack */
		if ((ACC_STATIC & access) == 0) {
			proxyMethod.visitVarInsn(ALOAD, localLoadPosition);
			localLoadPosition++;
			isVirtual = true;
		}
		insertTracingStatements(proxyMethod, className, logClassName,
				methodName, methodDescriptor, idMap, uniqueObjectPosition,
				isVirtual);
		Type args[] = Type.getArgumentTypes(methodDescriptor);

		for (int i = 0; i < args.length; i++) {
			int opcode = args[i].getOpcode(ILOAD);
			proxyMethod.visitVarInsn(opcode, localLoadPosition);
			if (args[i].equals(Type.DOUBLE_TYPE)
					|| args[i].equals(Type.LONG_TYPE)) {
				localLoadPosition += 2;
			} else {
				localLoadPosition++;
			}
		}
		if (isVirtual) {
			/* INVOKESPECIAL !!! because call should stay in class */
			proxyMethod.visitMethodInsn(INVOKESPECIAL, className, methodName,
					methodDescriptor);
		} else {
			proxyMethod.visitMethodInsn(INVOKESTATIC, className, methodName,
					methodDescriptor);
		}
		Type returnType = Type.getReturnType(methodDescriptor);
		int returnOpcode = returnType.getOpcode(IRETURN);
		proxyMethod.visitInsn(returnOpcode);
	}

	/**
	 * Insert tracing code
	 *
	 * @param proxyMethod
	 *            method visitor to which is used to insert the code
	 * @param className
	 *            Name of the class
	 * @param methodName
	 *            Name of the method
	 * @param access
	 *            access modifier
	 * @param methodDescriptor
	 *            Method descriptor
	 * @param idMap
	 *            idMap for reading the IDs to be inserted
	 */
	public static void insertLog(MethodVisitor proxyMethod, String className,
			String methodName, int access, String methodDescriptor,
			NameToIDMap idMap) {
		insertLog(proxyMethod, className, className, methodName, access,
				methodDescriptor, idMap);
	}

	/**
	 * Return the local variable position of the UniqueObject (always last
	 * argument)
	 *
	 * @param access
	 *            Access Modifier
	 * @param desc
	 *            Method descriptor
	 * @return
	 */
	public static int getUniqueObjectPosition(int access, String desc) {
		Type args[] = Type.getArgumentTypes(desc);
		int position = 0;
		if ((access & ACC_STATIC) == 0) {
			position++;
		}
		for (int i = 0; i < args.length; i++) {
			position += args[i].getSize();
		}
		return position;
	}

}
