package org.softevo.apibirthmark.asmadapters.calls;

import java.util.Vector;

import org.objectweb.asm.MethodAdapter;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;
import org.objectweb.asm.commons.LocalVariablesSorter;

import org.softevo.apibirthmark.constants.Constants;

import org.softevo.apibirthmark.asmadapters.calls.observed.ObservedCalls;
import org.softevo.apibirthmark.asmadapters.proxies.GenerateProxyClassAdapter;

import org.softevo.apibirthmark.tracing.mapping.ClassMethodID;
import org.softevo.apibirthmark.tracing.mapping.NameToIDMap;

/**
 * ASM {@link MethodAdapter} to replace calls to API methods with calls to the
 * corresponding proxy methods in a constructor. Special care must be taken
 * since their can be Invocations of API methods before the super constructor is
 * called, and therefore the objectID might not be initialized.
 *
 * @author schuler
 *
 */
public class InsertProxyCallsInConstructorMethodAdapter extends
		LocalVariablesSorter implements Opcodes {

	private String className;

	private Vector<String> precedingConstructorVector;

	private int classID;

	private int methodID;

	private boolean objectIDinitialized;

	private boolean waitForSuperOrThisConstructor;

	private ObservedCalls observedCalls;

	/**
	 * @param mv
	 *            MethodVisitor to forward the calls
	 * @param className
	 *            Name of the instrumented class
	 * @param access
	 *            Access modifier of this method
	 * @param methodName
	 *            Name of the method
	 * @param methodDescription
	 *            Descriptor of the method
	 * @param observedCalls
	 *            Used to decide if a call should be instrumented or not
	 * @param idMap
	 *            IDMap to insert the method that is instrumented
	 */
	public InsertProxyCallsInConstructorMethodAdapter(MethodVisitor mv,
			String className, int access, String methodName,
			String methodDescription, ObservedCalls observedCalls,
			NameToIDMap idMap) {
		super(access, methodDescription, mv);
		this.className = className;
		this.observedCalls = observedCalls;
		ClassMethodID classMethodID = idMap.getAndInsertClassMethodID(
				className, methodName, methodDescription);
		classID = classMethodID.getClassID();
		methodID = classMethodID.getMethodID();
		objectIDinitialized = false;
		waitForSuperOrThisConstructor = true;
		precedingConstructorVector = new Vector<String>();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitMethodInsn(int,
	 *      java.lang.String, java.lang.String, java.lang.String)
	 */
	public void visitMethodInsn(int opcode, String owner, String name,
			String desc) {
		String newName = name;
		Type returnType = Type.getReturnType(desc);
		boolean insertCall = false;
		if (name.indexOf("init>") == -1
				&& isObservedMethodCall(owner, name, desc)) {
			insertCall = true;
			newName = GenerateProxyClassAdapter.PREFIX + name;
		}

		else if (name.indexOf("<init>") != -1
				&& isObservedConstructor(owner, desc)) {// constructor
			insertCall = true;
			returnType = Type.VOID_TYPE;
		}
		if (insertCall) {
			mv.visitTypeInsn(NEW, "java/lang/MyUniqueObject");
			mv.visitInsn(DUP);
			mv.visitLdcInsn(new Integer(classID));
			mv.visitLdcInsn(new Integer(methodID));

			/*
			 * Before super constructor returned OBJECTID is not initialized !!!
			 */
			if (objectIDinitialized) {
				mv.visitVarInsn(ALOAD, getObjectPosition());
				mv.visitFieldInsn(GETFIELD, className,
						GenerateProxyClassAdapter.OBJECTID, "I");
			} else {
				mv.visitIntInsn(SIPUSH, Constants.CONSTRUCTORID);
			}
			mv.visitMethodInsn(INVOKESPECIAL, "java/lang/MyUniqueObject",
					"<init>", "(III)V");
			Type[] argumentTypes = Type.getArgumentTypes(desc);
			Type[] newArguments = new Type[argumentTypes.length + 1];
			System.arraycopy(argumentTypes, 0, newArguments, 0,
					argumentTypes.length);
			newArguments[newArguments.length - 1] = GenerateProxyClassAdapter.UNIQUETYPE;
			desc = Type.getMethodDescriptor(returnType, newArguments);
		}

		super.visitMethodInsn(opcode, owner, newName, desc);

		if (waitForSuperOrThisConstructor && opcode == INVOKESPECIAL) {
			if (isPrecedingConstructor(owner)) {
				removePrecedingConstructor(owner);
			} else {
				objectIDinitialized = true;
				waitForSuperOrThisConstructor = false;
			}
		}
	}

	/**
	 * The this reference is expected to be at position 0,but this can be
	 * changed by obfuscation methods
	 *
	 * @return local position of "this" reference
	 */
	protected int getObjectPosition() {
		return 0;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitTypeInsn(int, java.lang.String)
	 */
	public void visitTypeInsn(int opcode, String desc) {
		if (waitForSuperOrThisConstructor && opcode == NEW) {
			insertPrecedingConstructor(desc);
		}
		mv.visitTypeInsn(opcode, desc);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitEnd()
	 */
	public void visitEnd() {
		mv.visitEnd();
		if (waitForSuperOrThisConstructor) {
			if (className.equals("java/lang/Object")) {
				return;
			}
			throw new RuntimeException(
					"super constructor not found\nClass name: " + className);
		}
	}

	/**
	 * @param desc
	 */
	private void insertPrecedingConstructor(String desc) {
		precedingConstructorVector.add(desc);
	}

	private void removePrecedingConstructor(String owner) {
		if (precedingConstructorVector.size() > 0) {
			if (precedingConstructorVector.lastElement().equals(owner)) {
				precedingConstructorVector.remove(precedingConstructorVector
						.size() - 1);
			} else {
				throw new RuntimeException(precedingConstructorVector
						.lastElement()
						+ " OWNER " + owner);
			}
		}
	}

	/**
	 * Returns rue if the constructor comes before the call to the super
	 * constructor
	 *
	 * @param owner
	 *            Class name of the constructor
	 * @return True if the constructor comes before the call to the super
	 *         constructor
	 */
	private boolean isPrecedingConstructor(String owner) {
		if (precedingConstructorVector.size() > 0) {
			if (!precedingConstructorVector.lastElement().equals(owner)) {
				throw new RuntimeException(precedingConstructorVector
						.lastElement()
						+ " OWNER " + owner);
			}
			return true;
		}
		return false;
	}

	private boolean isObservedConstructor(String owner, String desc) {
		return observedCalls.isObservedConstructor(owner, desc);
	}

	private boolean isObservedMethodCall(String owner, String name, String desc) {
		return observedCalls.isObservedMethod(owner, name, desc);
	}

}
