package org.softevo.apibirthmark.asmadapters.calls;

import org.objectweb.asm.MethodAdapter;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;

import org.softevo.apibirthmark.constants.Constants;

import org.softevo.apibirthmark.asmadapters.BytecodeTasks;
import org.softevo.apibirthmark.asmadapters.calls.observed.ObservedCalls;
import org.softevo.apibirthmark.asmadapters.proxies.GenerateProxyClassAdapter;

import org.softevo.apibirthmark.tracing.mapping.ClassMethodID;
import org.softevo.apibirthmark.tracing.mapping.NameToIDMap;

/**
 * ASM {@link MethodAdapter} to replace calls to API methods with calls to the
 * corresponding proxy methods.
 *
 * @author schuler
 *
 */
public class InsertProxyCallsMethodAdapter extends MethodAdapter implements
		Opcodes {

	private String className;

	private int classID;

	private int methodID;

	private boolean objectMethod = false;

	private ObservedCalls observedCalls;

	/**
	 * Creates new instance of InsertProxyCallsMethodAdapter
	 *
	 * @param mv
	 *            MethodVisitor to forward the calls
	 * @param className
	 *            Name of the class that is instrumented
	 * @param access
	 *            Access modifiers of the class
	 * @param methodName
	 *            Name of the method that is instrumented
	 * @param methodDescription
	 *            Descriptor of the method
	 * @param observedCalls
	 *            Used to decide if a call should be instrumented or not
	 * @param idMap
	 *            IDMap to insert the method that is instrumented
	 */
	public InsertProxyCallsMethodAdapter(MethodVisitor mv, String className,
			int access, String methodName, String methodDescription,
			ObservedCalls observedCalls, NameToIDMap idMap) {
		super(mv);
		this.className = className;
		this.observedCalls = observedCalls;
		if ((access & ACC_STATIC) != 0) {
			objectMethod = false;
			// start Tracing if a static method is visited in observed Program
			mv.visitMethodInsn(INVOKESTATIC, BytecodeTasks.TRACER_CLASS , "start", "()V");

		} else {
			objectMethod = true;
		}
		ClassMethodID classMethodID = idMap.getAndInsertClassMethodID(
				className, methodName, methodDescription);
		classID = classMethodID.getClassID();
		methodID = classMethodID.getMethodID();

	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitMethodInsn(int,
	 *      java.lang.String, java.lang.String, java.lang.String)
	 */
	public void visitMethodInsn(int opcode, String owner, String name,
			String desc) {
		String newName = name;
		Type returnType = Type.getReturnType(desc);
		boolean insertCall = false;
		if (name.indexOf("init>") == -1
				&& isObservedMethodCall(owner, name, desc)) {
			insertCall = true;
			newName = GenerateProxyClassAdapter.getModifiedName(name);
		} else if (name.indexOf("<init>") != -1
				&& isObservedConstructor(owner, desc)) {// constructor
			insertCall = true;
			returnType = Type.VOID_TYPE;
		}
		if (insertCall) {
			// Push unique object on stack
			mv.visitTypeInsn(NEW, GenerateProxyClassAdapter.UNIQUEOBJECT);
			mv.visitInsn(DUP);
			mv.visitLdcInsn(new Integer(classID));
			mv.visitLdcInsn(new Integer(methodID));
			if (objectMethod) {
				mv.visitVarInsn(ALOAD, 0);
				mv.visitFieldInsn(GETFIELD, className,
						GenerateProxyClassAdapter.OBJECTID, "I");
			} else {
				mv.visitIntInsn(SIPUSH, Constants.STATICOBJECTID);
			}
			mv.visitMethodInsn(INVOKESPECIAL,
					GenerateProxyClassAdapter.UNIQUEOBJECT, "<init>", "(III)V");
			Type[] argumentTypes = Type.getArgumentTypes(desc);
			Type[] newArguments = new Type[argumentTypes.length + 1];
			System.arraycopy(argumentTypes, 0, newArguments, 0,
					argumentTypes.length);
			newArguments[newArguments.length - 1] = GenerateProxyClassAdapter.UNIQUETYPE;
			desc = Type.getMethodDescriptor(returnType, newArguments);
		}
		super.visitMethodInsn(opcode, owner, newName, desc);
	}

	private boolean isObservedConstructor(String owner, String desc) {
		return observedCalls.isObservedConstructor(owner, desc);
	}

	private boolean isObservedMethodCall(String owner, String name, String desc) {
		return observedCalls.isObservedMethod(owner, name, desc);
	}

}
