package org.softevo.apibirthmark.asmadapters.proxies;

import java.util.HashMap;
import java.util.Vector;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.objectweb.asm.AnnotationVisitor;
import org.objectweb.asm.Attribute;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodAdapter;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;

import org.softevo.apibirthmark.asmadapters.BytecodeTasks;

import org.softevo.apibirthmark.tracing.mapping.NameToIDMap;

/**
 * MethodAdapter used to generate proxy constructors. Here the code is copied
 * and tracing code and special code to identify objects is inserted
 *
 * @author schuler
 *
 */
public class GenerateConstructorProxyMethodAdapter extends MethodAdapter
		implements Opcodes {

	private static Logger logger = Logger
			.getLogger(GenerateConstructorProxyMethodAdapter.class);

	static {
		logger.setLevel(Level.OFF);
	}

	private MethodVisitor proxyMethod;

	private String className;

	private String methodName;

	private HashMap<Label, Label> labelMap = new HashMap<Label, Label>();

	private String methodDescriptor;

	private NameToIDMap idMap;

	private int access;

	private int additionalArguments = 1;

	private int numberOfArguments;

	private Vector<String> precedingConstructorVector = new Vector<String>();

	private boolean waitForSuperOrThisConstructor;

	private boolean objectSubclass;

	/**
	 * gives new GenerateConstructorProxyMethodAdapter
	 *
	 * @param mv
	 *            original MethodVisitor
	 * @param proxyMethod
	 *            MethodVisitor for proxy method
	 * @param className
	 *            name of the class that is instrumented
	 * @param access
	 *            access modifier of the method that is instrumented
	 * @param name
	 *            name of the method that is instrumented
	 * @param methodDescriptor
	 *            method descriptor of the original method
	 * @param objectSubclass
	 *            indicate if this class is a direct subclass of
	 *            java/lang/object ( in this case objectIDs are inserted)
	 * @param idMap
	 *            idMap that contains the method IDs
	 */
	public GenerateConstructorProxyMethodAdapter(MethodVisitor mv,
			MethodVisitor proxyMethod, String className, int access,
			String name, String methodDescriptor, boolean objectSubclass,
			NameToIDMap idMap) {
		super(mv);
		this.proxyMethod = proxyMethod;
		this.className = className;
		this.methodName = name;
		this.methodDescriptor = methodDescriptor;
		this.objectSubclass = objectSubclass;
		this.idMap = idMap;
		this.access = access;
		waitForSuperOrThisConstructor = true;
		numberOfArguments = BytecodeTasks.getUniqueObjectPosition(access,
				methodDescriptor);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitCode()
	 */
	public void visitCode() {
		proxyMethod.visitCode();
		mv.visitCode();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitEnd()
	 */
	public void visitEnd() {
		proxyMethod.visitEnd();
		mv.visitEnd();
		if (waitForSuperOrThisConstructor) {
			throw new RuntimeException(
					"super constructor not found\nClass name: " + className
							+ " " + methodName);
		}
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitFieldInsn(int,
	 *      java.lang.String, java.lang.String, java.lang.String)
	 */
	public void visitFieldInsn(int opcode, String owner, String name,
			String desc) {
		proxyMethod.visitFieldInsn(opcode, owner, name, desc);
		mv.visitFieldInsn(opcode, owner, name, desc);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitIincInsn(int, int)
	 */
	public void visitIincInsn(int var, int increment) {
		int newVar = var;
		if (var >= numberOfArguments) {
			newVar += additionalArguments;
		}
		proxyMethod.visitIincInsn(newVar, increment);
		mv.visitIincInsn(var, increment);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitInsn(int)
	 */
	public void visitInsn(int opcode) {
		proxyMethod.visitInsn(opcode);
		mv.visitInsn(opcode);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitIntInsn(int, int)
	 */
	public void visitIntInsn(int opcode, int operand) {
		proxyMethod.visitIntInsn(opcode, operand);
		mv.visitIntInsn(opcode, operand);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitJumpInsn(int,
	 *      org.objectweb.asm.Label)
	 */
	public void visitJumpInsn(int opcode, Label label) {
		proxyMethod.visitJumpInsn(opcode, getLabel(label));
		mv.visitJumpInsn(opcode, label);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitLabel(org.objectweb.asm.Label)
	 */
	public void visitLabel(Label label) {
		proxyMethod.visitLabel(getLabel(label));
		mv.visitLabel(label);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitLdcInsn(java.lang.Object)
	 */
	public void visitLdcInsn(Object cst) {
		proxyMethod.visitLdcInsn(cst);
		mv.visitLdcInsn(cst);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitLineNumber(int,
	 *      org.objectweb.asm.Label)
	 */
	public void visitLineNumber(int line, Label start) {
		proxyMethod.visitLineNumber(line, getLabel(start));
		mv.visitLineNumber(line, start);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitLocalVariable(java.lang.String,
	 *      java.lang.String, java.lang.String, org.objectweb.asm.Label,
	 *      org.objectweb.asm.Label, int)
	 */
	public void visitLocalVariable(String name, String desc, String signature,
			Label start, Label end, int index) {
		proxyMethod.visitLocalVariable(name, desc, signature, getLabel(start),
				getLabel(end), index);
		mv.visitLocalVariable(name, desc, signature, start, end, index);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitLookupSwitchInsn(org.objectweb.asm.Label,
	 *      int[], org.objectweb.asm.Label[])
	 */
	public void visitLookupSwitchInsn(Label dflt, int[] keys, Label[] labels) {
		Label[] newLabels = new Label[labels.length];
		for (int i = 0; i < newLabels.length; i++) {
			newLabels[i] = getLabel(labels[i]);
		}
		proxyMethod.visitLookupSwitchInsn(getLabel(dflt), keys, newLabels);
		mv.visitLookupSwitchInsn(dflt, keys, labels);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitMaxs(int, int)
	 */
	public void visitMaxs(int maxStack, int maxLocals) {
		proxyMethod.visitMaxs(maxStack, maxLocals);
		mv.visitMaxs(maxStack, maxLocals);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitMethodInsn(int,
	 *      java.lang.String, java.lang.String, java.lang.String)
	 */
	public void visitMethodInsn(int opcode, String owner, String name,
			String desc) {
		boolean isSuperOrThisConstructor = false;
		if (waitForSuperOrThisConstructor && opcode == INVOKESPECIAL) {
			if (isPrecedingConstructor(owner)) {
				removePrecedingConstructor(owner);
			} else {
				isSuperOrThisConstructor = true;
			}

		}

		proxyMethod.visitMethodInsn(opcode, owner, name, desc);
		mv.visitMethodInsn(opcode, owner, name, desc);

		if (isSuperOrThisConstructor) {
			waitForSuperOrThisConstructor = false;
			if (objectSubclass) {
				insertDynamicObjectID(proxyMethod, className);
				insertDynamicObjectID(mv, className);
			}
			BytecodeTasks.insertTracingStatements(proxyMethod, className, name,
					methodDescriptor, idMap, BytecodeTasks
							.getUniqueObjectPosition(access, methodDescriptor),
					true);

		}
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitMultiANewArrayInsn(java.lang.String,
	 *      int)
	 */
	public void visitMultiANewArrayInsn(String desc, int dims) {
		proxyMethod.visitMultiANewArrayInsn(desc, dims);
		mv.visitMultiANewArrayInsn(desc, dims);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitTableSwitchInsn(int, int,
	 *      org.objectweb.asm.Label, org.objectweb.asm.Label[])
	 */
	public void visitTableSwitchInsn(int min, int max, Label dflt,
			Label[] labels) {
		Label[] newLabels = new Label[labels.length];
		for (int i = 0; i < newLabels.length; i++) {
			newLabels[i] = getLabel(labels[i]);
		}
		proxyMethod.visitTableSwitchInsn(min, max, getLabel(dflt), newLabels);
		mv.visitTableSwitchInsn(min, max, dflt, labels);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitTryCatchBlock(org.objectweb.asm.Label,
	 *      org.objectweb.asm.Label, org.objectweb.asm.Label, java.lang.String)
	 */
	public void visitTryCatchBlock(Label start, Label end, Label handler,
			String type) {
		proxyMethod.visitTryCatchBlock(getLabel(start), getLabel(end),
				getLabel(handler), type);
		mv.visitTryCatchBlock(start, end, handler, type);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitTypeInsn(int, java.lang.String)
	 */
	public void visitTypeInsn(int opcode, String desc) {
		if (waitForSuperOrThisConstructor && opcode == NEW) {
			insertPrecedingConstructor(desc);
		}
		proxyMethod.visitTypeInsn(opcode, desc);
		mv.visitTypeInsn(opcode, desc);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitVarInsn(int, int)
	 */
	public void visitVarInsn(int opcode, int var) {
		int newVar = var;
		if (var >= numberOfArguments) {
			newVar += additionalArguments;
		}
		proxyMethod.visitVarInsn(opcode, newVar);
		mv.visitVarInsn(opcode, var);
	}

	/**
	 * Inserts code to initialize the object ID file
	 *
	 * @param insertMV
	 *            method visitor to which the code should be added
	 * @param className
	 *            name of the class
	 */
	private static void insertDynamicObjectID(MethodVisitor insertMV,
			String className) {
		insertMV.visitVarInsn(ALOAD, 0);
		insertMV.visitMethodInsn(INVOKESTATIC, "java/lang/MyUniqueObject",
				"getID", "()I");
		insertMV.visitFieldInsn(PUTFIELD, className,
				GenerateProxyClassAdapter.OBJECTID, "I");
		insertMV.visitFieldInsn(GETSTATIC, className,
				GenerateProxyClassAdapter.STATICOBJECTID, "I");
		insertMV.visitInsn(ICONST_1);
		insertMV.visitInsn(IADD);
		insertMV.visitFieldInsn(PUTSTATIC, className,
				GenerateProxyClassAdapter.STATICOBJECTID, "I");
	}

	/**
	 * Used for recomputation of labels
	 *
	 * @param label
	 * @return
	 */
	private Label getLabel(Label label) {
		if (labelMap.containsKey(label)) {
			return (Label) labelMap.get(label);
		}
		Label returnLabel = new Label();
		labelMap.put(label, returnLabel);
		return returnLabel;
	}

	/**
	 * called when a NEW instruction is seen before the super constructor was
	 * called
	 *
	 * @param desc
	 */
	private void insertPrecedingConstructor(String desc) {
		precedingConstructorVector.add(desc);
	}

	/**
	 * called when an constructor is called that is not the super constructor
	 *
	 * @param owner
	 */
	private void removePrecedingConstructor(String owner) {
		if (precedingConstructorVector.size() > 0) {
			if (precedingConstructorVector.lastElement().equals(owner)) {
				precedingConstructorVector.remove(precedingConstructorVector
						.size() - 1);
			} else {
				throw new RuntimeException(precedingConstructorVector
						.lastElement()
						+ " OWNER " + owner);
			}
		}

	}

	/**
	 * Checks if the constructor is a constructor preceding the super
	 * constructor
	 *
	 * @param owner
	 *            name of the constructor's class
	 * @return
	 */
	private boolean isPrecedingConstructor(String owner) {
		if (precedingConstructorVector.size() > 0) {
			if (!precedingConstructorVector.lastElement().equals(owner)) {
				throw new RuntimeException(precedingConstructorVector
						.lastElement()
						+ " OWNER " + owner);
			}
			return true;
		}
		return false;
	}

	/* (non-Javadoc)
	 * @see org.objectweb.asm.MethodAdapter#visitAnnotation(java.lang.String, boolean)
	 */
	@Override
	public AnnotationVisitor visitAnnotation(String desc, boolean visible) {
		logger.info(desc);
		return super.visitAnnotation(desc, visible);
	}

	/* (non-Javadoc)
	 * @see org.objectweb.asm.MethodAdapter#visitAnnotationDefault()
	 */
	@Override
	public AnnotationVisitor visitAnnotationDefault() {
		logger.info("visitAnnotationDefault() {");
		return super.visitAnnotationDefault();
	}

	/* (non-Javadoc)
	 * @see org.objectweb.asm.MethodAdapter#visitAttribute(org.objectweb.asm.Attribute)
	 */
	@Override
	public void visitAttribute(Attribute attr) {
		logger.info(attr);
		super.visitAttribute(attr);
	}

	/* (non-Javadoc)
	 * @see org.objectweb.asm.MethodAdapter#visitParameterAnnotation(int, java.lang.String, boolean)
	 */
	@Override
	public AnnotationVisitor visitParameterAnnotation(int parameter,
			String desc, boolean visible) {
		logger.info(desc);
		return super.visitParameterAnnotation(parameter, desc, visible);
	}

}
