package org.softevo.apibirthmark.asmadapters.proxies;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.FieldVisitor;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;
import org.objectweb.asm.util.CheckMethodAdapter;

import org.softevo.apibirthmark.constants.Constants;

import org.softevo.apibirthmark.asmadapters.BytecodeTasks;

import org.softevo.apibirthmark.tracing.mapping.NameToIDMap;

/**
 * ClassAdapter for instrumenting the API.For every Method and Constructor a
 * proxy method is inserted
 *
 * @author schuler
 *
 */
public class GenerateProxyClassAdapter extends ClassAdapter implements Opcodes {

	private static Logger logger = Logger
			.getLogger(GenerateProxyClassAdapter.class);

	static {
		logger.setLevel(Level.ALL);
	}

	/**
	 * Prefix to avoid name clashes
	 */
	public static final String PREFIX = "XYZ";

	/**
	 * Name of the field that holds the dynamically computed objectID
	 */
	public static final String OBJECTID = PREFIX + "objectID";

	/**
	 * Name of the field that counts the number of instantiated objects
	 */
	public static final String STATICOBJECTID = PREFIX + "StaticObjectID";

	/**
	 * Name of the unique object (used to make method signatures unique)
	 */
	public static final String UNIQUEOBJECT = "java/lang/MyUniqueObject";

	/**
	 * Type of the unique object
	 */
	public static final Type UNIQUETYPE = Type
			.getType("L" + UNIQUEOBJECT + ";");



	private String className;

	private boolean isExluded;

	private int count;

	private NameToIDMap idMap;

	private int classAccess;

	private boolean insertObjectID;

	/**
	 * generates new GenerateProxyClassAdapter. Inserts proxy methods in this
	 * class
	 *
	 * @param cv
	 *            ClassVisitor of ASM chain
	 * @param idMap
	 *            computed class and method IDs are inserted into this map and
	 *            their corresponding names
	 */
	public GenerateProxyClassAdapter(ClassVisitor cv, NameToIDMap idMap) {
		super(cv);
		this.idMap = idMap;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.ClassAdapter#visit(int, int, java.lang.String,
	 *      java.lang.String, java.lang.String, java.lang.String[])
	 */
	public void visit(int version, int access, String name, String signature,
			String superName, String[] interfaces) {
		super.visit(version, access, name, signature, superName, interfaces);
		className = name;
		classAccess = access;
		isExluded = isExluded(className);
		if (isExluded) {
			logger.info("Excluding Class:" + className);
		}
		if (idMap != null) {
			if (!idMap.contains(className)) {
				idMap.insertClass(className);
			}
		} else {
			throw new RuntimeException("No IDMap");
		}
		if (superName != null
				&& (superName.equals("java/lang/Object") || superName
						.equals("java/lang/Number"))) {
			insertObjectID = true;
		}
		count = 0;
	}

	/**
	 * Insert proxy methods
	 *
	 * @see org.objectweb.asm.ClassAdapter#visitMethod(int, java.lang.String,
	 *      java.lang.String, java.lang.String, java.lang.String[])
	 */
	public MethodVisitor visitMethod(int access, String name, String desc,
			String signature, String[] exceptions) {
		MethodVisitor mv = super.visitMethod(access, name, desc, signature,
				exceptions);
		if (isExluded) {
			return mv;
		}
		if (!idMap.containsMethod(className, name, desc)) {
			idMap.insertMethod(className, name, desc);
		}

		String proxySignature = getProxySignature(signature);
		MethodVisitor proxyMethod;
		if (name.equals("<init>")) {
			if (mv != null) {
				Type[] args = Type.getArgumentTypes(desc);
				Type[] newArgs = appendUniqueType(args);
				Type returnType = Type.VOID_TYPE;
				proxyMethod = cv.visitMethod(access, name, Type
						.getMethodDescriptor(returnType, newArgs),
						proxySignature, exceptions);
				return new CheckMethodAdapter(
						new GenerateConstructorProxyMethodAdapter(mv,
								proxyMethod, className, access, name, desc,
								insertObjectID, idMap));
			}
		} else if (name.equals("<clinit>")) {
			// do nothing for class initializers
		} else {
			if (mv != null) {
				count++;
				if ((access & ACC_NATIVE) != 0) {
					proxyMethod = cv.visitMethod(access ^ ACC_NATIVE, PREFIX
							+ name, getProxyDescriptor(desc), proxySignature,
							exceptions);
					/* insert Code into proxy method (native has none) */
					proxyMethod.visitCode();
					BytecodeTasks.insertLog(proxyMethod, className, name,
							access, desc, idMap);
					proxyMethod.visitMaxs(0, 0);
					proxyMethod.visitEnd();
					return mv;
				}
				/* instrumentation for abstract methods */
				else if ((access & ACC_ABSTRACT) != 0) {
					proxyMethod = cv.visitMethod(access, PREFIX + name,
							getProxyDescriptor(desc), proxySignature,
							exceptions);
					return mv;
				} else {
					proxyMethod = cv.visitMethod(access, PREFIX + name,
							getProxyDescriptor(desc), proxySignature,
							exceptions);
					/*
					 * special case for java/lang/Class since methods do depend
					 * on the depth of the method stack
					 */
					if (className.equals("java/lang/Class")) {
						return new GenerateCopyProxyMethodAdapter(mv, access,
								desc, proxyMethod, className, name, idMap);

					}
					return getDefaultAdapter(mv, proxyMethod, className,
							access, name, desc, signature, exceptions, idMap);
				}
			}
		}
		return mv;
	}

	/**
	 * Computes signature of the proxy method for given original signature.
	 *
	 * @param signature
	 * @return
	 */
	private String getProxySignature(String signature) {
		if (signature == null) {
			return null;
		}
		StringBuffer newSignature = new StringBuffer();
		newSignature.append(signature.substring(0, signature.lastIndexOf(')')));
		newSignature.append(UNIQUETYPE.toString());
		newSignature.append(signature.substring(signature.lastIndexOf(')')));
		return newSignature.toString();
	}

	/**
	 * Default adapter that is used for proxy methods
	 *
	 * @return
	 */
	protected MethodVisitor getDefaultAdapter(MethodVisitor mv,
			MethodVisitor proxyMethod, String className2, int access,
			String name, String desc, String signature, String[] exceptions,
			NameToIDMap idMap2) {
		return new CheckMethodAdapter(new GenerateProxyMethodAdapter(mv,
				proxyMethod, className2, access, name, desc, signature,
				exceptions, idMap2));
	}

	/**
	 * Checks if the class should be excluded from instrumentation.
	 *
	 * @param checkClassName
	 *            class name to check
	 * @return
	 */
	private boolean isExluded(String checkClassName) {
		for (int i = 0; i < Constants.EXCLUDE_CLASSES.length; i++) {
			if (checkClassName.equals(Constants.EXCLUDE_CLASSES[i])) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Return a copy of given Type array with the UniqueType appended
	 *
	 * @param args
	 *            Type array where UniqueType should be appended
	 * @return
	 */
	private Type[] appendUniqueType(Type[] args) {
		Type[] newArgs = new Type[args.length + 1];
		System.arraycopy(args, 0, newArgs, 0, args.length);
		newArgs[newArgs.length - 1] = UNIQUETYPE;
		return newArgs;
	}

	/**
	 * Return method descriptor for proxy method for given original descriptor
	 *
	 * @param desc
	 *            method descriptor of the method to be proxied
	 * @return method descriptor of the proxy method
	 *
	 */
	private String getProxyDescriptor(String desc) {
		Type[] args = Type.getArgumentTypes(desc);
		Type[] newArgs = appendUniqueType(args);
		Type returnType = Type.getReturnType(desc);
		return Type.getMethodDescriptor(returnType, newArgs);

	}

	/**
	 * Insert field that holds the ObjectID
	 *
	 * @see org.objectweb.asm.ClassAdapter#visitEnd()
	 */
	public void visitEnd() {
		if (!isExluded(className)) {
			int modifier = ACC_PUBLIC;
			/* DO NOTHING FOR INTERFACES (could lead to multiple ObjectIDs */
			if (!((classAccess & ACC_INTERFACE) != 0)) {
				if (insertObjectID) {
					FieldVisitor fv1 = visitField(modifier, OBJECTID, "I",
							null, null);
					fv1.visitEnd();
					FieldVisitor fv2 = visitField(ACC_STATIC + ACC_PRIVATE,
							STATICOBJECTID, "I", null, null);
					fv2.visitEnd();
				}
			}
		} else {
			logger.info("Class excluded from instrumentation: " + className);
		}
		super.visitEnd();
	}

	/**
	 * Get name of proxy method for given method name
	 *
	 * @param name
	 *            name of the original method for which a proxy method is
	 *            introduced
	 * @return name of proxy method
	 */
	public static String getModifiedName(String name) {
		return GenerateProxyClassAdapter.PREFIX + name;
	}

}
