package org.softevo.apibirthmark.asmadapters.proxies;

import org.objectweb.asm.AnnotationVisitor;
import org.objectweb.asm.Attribute;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodAdapter;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;

import org.softevo.apibirthmark.asmadapters.BytecodeTasks;

import org.softevo.apibirthmark.tracing.mapping.NameToIDMap;

/**
 * Generate A proxy method by inserting the tracing code and then calling the
 * original method that is proxied. If the original method returns an argument,
 * this argument is returned by the proxy method.
 *
 * @author schuler
 *
 */
public class GenerateProxyMethodAdapter extends MethodAdapter implements
		Opcodes {

	private MethodVisitor proxyMethod;

	private int access;

	private String methodName;

	private String methodDescriptor;

	protected String className;

	private NameToIDMap idMap;

	protected String logClassName;

	/**
	 * Create a new GenerateProxyMethodAdapter
	 *
	 * @param mv
	 *            MethodVisitor that represents the original method
	 * @param proxyMethod
	 *            MethodVisitor that represents the proxy method
	 * @param className
	 *            Name of the class
	 * @param access
	 *            Access modifier of the method
	 * @param methodName
	 *            The method's name
	 * @param desc
	 *            Method descriptor
	 * @param signature
	 *            Signature of the method
	 * @param exceptions
	 *            Exceptions thrown by this method
	 * @param idMap
	 *            NameToIDMap that holds the IDs for classes and methods
	 */
	public GenerateProxyMethodAdapter(MethodVisitor mv,
			MethodVisitor proxyMethod, String className, int access,
			String methodName, String desc, String signature,
			String[] exceptions, NameToIDMap idMap) {
		super(mv);
		this.proxyMethod = proxyMethod;
		this.className = className;
		this.access = access;
		this.methodName = methodName;
		this.methodDescriptor = desc;
		this.idMap = idMap;
		this.logClassName = className;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitCode()
	 */
	public void visitCode() {
		mv.visitCode();
		BytecodeTasks.insertLog(proxyMethod, className, logClassName,
				methodName, access, methodDescriptor, idMap);
		proxyMethod.visitCode();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitAnnotation(java.lang.String,
	 *      boolean)
	 */
	public AnnotationVisitor visitAnnotation(String desc, boolean visible) {
		proxyMethod.visitAnnotation(desc, visible);
		return super.visitAnnotation(desc, visible);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitAnnotationDefault()
	 */
	public AnnotationVisitor visitAnnotationDefault() {
		proxyMethod.visitAnnotationDefault();
		return super.visitAnnotationDefault();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitAttribute(org.objectweb.asm.Attribute)
	 */
	public void visitAttribute(Attribute attr) {
		proxyMethod.visitAttribute(attr);
		super.visitAttribute(attr);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitEnd()
	 */
	public void visitEnd() {
		proxyMethod.visitEnd();
		super.visitEnd();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitMaxs(int, int)
	 */
	public void visitMaxs(int maxStack, int maxLocals) {
		proxyMethod.visitMaxs(maxStack, maxLocals);
		super.visitMaxs(maxStack, maxLocals);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitParameterAnnotation(int,
	 *      java.lang.String, boolean)
	 */
	public AnnotationVisitor visitParameterAnnotation(int parameter,
			String desc, boolean visible) {
		proxyMethod.visitParameterAnnotation(parameter, desc, visible);
		return super.visitParameterAnnotation(parameter, desc, visible);
	}

	public void visitLineNumber(int line, Label start) {
		super.visitLineNumber(line, start);
	}
}