package org.softevo.apibirthmark.asmadapters.proxies;

import java.util.HashMap;

import org.objectweb.asm.AnnotationVisitor;
import org.objectweb.asm.Attribute;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodAdapter;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;

import org.softevo.apibirthmark.asmadapters.BytecodeTasks;

/**
 * A MethodAdapter that delegates calls to a proxy method (such that the code of
 * the original method is also contained in the proxy method), and does the
 * necessary calculations for labels, and new arguments.
 *
 */

public class ProxyMethodAdapter extends MethodAdapter implements Opcodes {

	protected MethodVisitor proxyMethod;

	private HashMap<Label, Label> labelMap;

	/**
	 * Stores the number of Arguments. Indicates the position of the additional
	 * argument of the proxy method that carries the caller information.
	 */
	protected int numberOfArguments;

	private int additionalArguments;

	public ProxyMethodAdapter(MethodVisitor mv, int mvAccess,
			String mvDescriptor, MethodVisitor proxyMethod,
			int additionalArguments) {
		super(mv);
		this.proxyMethod = proxyMethod;
		labelMap = new HashMap<Label, Label>();
		numberOfArguments = BytecodeTasks.getUniqueObjectPosition(mvAccess,
				mvDescriptor);
		this.additionalArguments = additionalArguments;
	}

	/**
	 * Returns the corresponding label of the proxy method for given label of the
	 * original method
	 *
	 * @param label
	 *            label of the original method
	 * @return corresponding label of the proxy method
	 */
	private Label getLabel(Label label) {
		if (labelMap.containsKey(label)) {
			return (Label) labelMap.get(label);
		}
		Label returnLabel = new Label();
		labelMap.put(label, returnLabel);
		return returnLabel;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitCode()
	 */
	public void visitCode() {
		proxyMethod.visitCode();
		mv.visitCode();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitEnd()
	 */
	public void visitEnd() {
		proxyMethod.visitEnd();
		mv.visitEnd();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitFieldInsn(int,
	 *      java.lang.String, java.lang.String, java.lang.String)
	 */
	public void visitFieldInsn(int opcode, String owner, String name,
			String desc) {
		proxyMethod.visitFieldInsn(opcode, owner, name, desc);
		mv.visitFieldInsn(opcode, owner, name, desc);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitIincInsn(int, int)
	 */
	public void visitIincInsn(int var, int increment) {
		int newVar = var;
		if (var >= numberOfArguments) {
			newVar += additionalArguments;
		}
		proxyMethod.visitIincInsn(newVar, increment);
		mv.visitIincInsn(var, increment);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitInsn(int)
	 */
	public void visitInsn(int opcode) {
		proxyMethod.visitInsn(opcode);
		mv.visitInsn(opcode);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitIntInsn(int, int)
	 */
	public void visitIntInsn(int opcode, int operand) {
		proxyMethod.visitIntInsn(opcode, operand);
		mv.visitIntInsn(opcode, operand);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitJumpInsn(int,
	 *      org.objectweb.asm.Label)
	 */
	public void visitJumpInsn(int opcode, Label label) {
		proxyMethod.visitJumpInsn(opcode, getLabel(label));
		mv.visitJumpInsn(opcode, label);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitLabel(org.objectweb.asm.Label)
	 */
	public void visitLabel(Label label) {
		proxyMethod.visitLabel(getLabel(label));
		mv.visitLabel(label);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitLdcInsn(java.lang.Object)
	 */
	public void visitLdcInsn(Object cst) {
		proxyMethod.visitLdcInsn(cst);
		mv.visitLdcInsn(cst);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitLineNumber(int,
	 *      org.objectweb.asm.Label)
	 */
	public void visitLineNumber(int line, Label start) {
		proxyMethod.visitLineNumber(line, getLabel(start));
		mv.visitLineNumber(line, start);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitLocalVariable(java.lang.String,
	 *      java.lang.String, java.lang.String, org.objectweb.asm.Label,
	 *      org.objectweb.asm.Label, int)
	 */
	public void visitLocalVariable(String name, String desc, String signature,
			Label start, Label end, int index) {
		proxyMethod.visitLocalVariable(name, desc, signature, getLabel(start),
				getLabel(end), index);
		mv.visitLocalVariable(name, desc, signature, start, end, index);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitLookupSwitchInsn(org.objectweb.asm.Label,
	 *      int[], org.objectweb.asm.Label[])
	 */
	public void visitLookupSwitchInsn(Label dflt, int[] keys, Label[] labels) {
		Label[] newLabels = new Label[labels.length];
		for (int i = 0; i < newLabels.length; i++) {
			newLabels[i] = getLabel(labels[i]);
		}
		proxyMethod.visitLookupSwitchInsn(getLabel(dflt), keys, newLabels);
		mv.visitLookupSwitchInsn(dflt, keys, labels);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitMaxs(int, int)
	 */
	public void visitMaxs(int maxStack, int maxLocals) {
		proxyMethod.visitMaxs(maxStack, maxLocals);
		mv.visitMaxs(maxStack, maxLocals);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitMethodInsn(int,
	 *      java.lang.String, java.lang.String, java.lang.String)
	 */
	public void visitMethodInsn(int opcode, String owner, String name,
			String desc) {
		proxyMethod.visitMethodInsn(opcode, owner, name, desc);
		mv.visitMethodInsn(opcode, owner, name, desc);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitMultiANewArrayInsn(java.lang.String,
	 *      int)
	 */
	public void visitMultiANewArrayInsn(String desc, int dims) {
		proxyMethod.visitMultiANewArrayInsn(desc, dims);
		mv.visitMultiANewArrayInsn(desc, dims);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitTableSwitchInsn(int, int,
	 *      org.objectweb.asm.Label, org.objectweb.asm.Label[])
	 */
	public void visitTableSwitchInsn(int min, int max, Label dflt,
			Label[] labels) {
		Label[] newLabels = new Label[labels.length];
		for (int i = 0; i < newLabels.length; i++) {
			newLabels[i] = getLabel(labels[i]);
		}
		proxyMethod.visitTableSwitchInsn(min, max, getLabel(dflt), newLabels);
		mv.visitTableSwitchInsn(min, max, dflt, labels);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitTryCatchBlock(org.objectweb.asm.Label,
	 *      org.objectweb.asm.Label, org.objectweb.asm.Label, java.lang.String)
	 */
	public void visitTryCatchBlock(Label start, Label end, Label handler,
			String type) {
		proxyMethod.visitTryCatchBlock(getLabel(start), getLabel(end),
				getLabel(handler), type);
		mv.visitTryCatchBlock(start, end, handler, type);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitTypeInsn(int, java.lang.String)
	 */
	public void visitTypeInsn(int opcode, String desc) {
		proxyMethod.visitTypeInsn(opcode, desc);
		mv.visitTypeInsn(opcode, desc);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitVarInsn(int, int)
	 */
	public void visitVarInsn(int opcode, int var) {
		int newVar = var;
		if (var >= numberOfArguments) {
			newVar += additionalArguments;
		}
		proxyMethod.visitVarInsn(opcode, newVar);
		mv.visitVarInsn(opcode, var);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitAnnotation(java.lang.String,
	 *      boolean)
	 */
	public AnnotationVisitor visitAnnotation(String desc, boolean visible) {
		proxyMethod.visitAnnotation(desc, visible);
		return super.visitAnnotation(desc, visible);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitAnnotationDefault()
	 */
	public AnnotationVisitor visitAnnotationDefault() {
		proxyMethod.visitAnnotationDefault();
		return super.visitAnnotationDefault();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitAttribute(org.objectweb.asm.Attribute)
	 */
	public void visitAttribute(Attribute attr) {
		proxyMethod.visitAttribute(attr);
		super.visitAttribute(attr);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.objectweb.asm.MethodAdapter#visitParameterAnnotation(int,
	 *      java.lang.String, boolean)
	 */
	public AnnotationVisitor visitParameterAnnotation(int parameter,
			String desc, boolean visible) {
		proxyMethod.visitParameterAnnotation(parameter, desc, visible);
		return super.visitParameterAnnotation(parameter, desc, visible);
	}

}
