package org.softevo.apibirthmark.tracing;

import java.io.File;
import java.io.FilenameFilter;
import java.io.PrintStream;
import java.io.Serializable;
import java.util.Arrays;


import org.softevo.apibirthmark.tracing.tracehandler.CalleeClassHandler;
import org.softevo.apibirthmark.tracing.tracehandler.CallerClassHandler;
import org.softevo.apibirthmark.tracing.tracehandler.CallerObjectHandler;
import org.softevo.apibirthmark.tracing.tracehandler.ExcludeHandler;
import org.softevo.apibirthmark.tracing.tracehandler.GlobalWindowTraceHandler;
import org.softevo.apibirthmark.tracing.tracehandler.ObjectTraceHandler;
import org.softevo.apibirthmark.tracing.tracehandler.TotalTraceHandler;
import org.softevo.apibirthmark.tracing.tracehandler.TraceObjectHandler;
import org.softevo.apibirthmark.tracing.tracehandler.ValidationHandler;
import org.softevo.apibirthmark.constants.Constants;

/**
 * This class is needed by the instrumented code at runtime. All observed API
 * methods call the traceINSTR method of the tracer. Then these observed calls
 * are further processed by this class
 *
 * @author schuler
 *
 */
public class Tracer {


	public static final String OBSERVED_TRACE = Constants.OUTPUT_DIRECTORY
			+ "validation/observedTrace.xml";

	public static final String NAMETOID_FILE_LOCATION = Constants.RUNTIME_IDMAP_LOCATION;

	/**
	 * Reference to System.out used when an application changes Sys.out
	 */
	public static final PrintStream mySysOut = System.out;

	/**
	 * Reference to System.err used when an application changes Sys.err
	 */
	public static final PrintStream mySysErr = System.err;

	private static int traceCount = 0;

	private static TraceObject traceObject;

	private static TraceObjectHandler traceObjectHandler = new ObjectTraceHandler();

	/**
	 * Location of the result file for the tracing results
	 */
	public static String resultFileLocation;

	/**
	 * Default Location for the result file
	 */
	private static final String DEFAULT_RESULT_FILE_LOCATION = Constants.OUTPUT_DIRECTORY
			+ "result.ser";

	private static boolean started = false;

	static {
		String traceHandler = System
				.getProperty(Constants.TRACE_HANDLER_PROPERTY);
		if (traceHandler != null) {
			if (traceHandler.equals(Constants.VALIDATION_TRACE_HANDLER)) {
				mySysErr.println("SETTING VALIDATION HANDLER");
				traceObjectHandler = new ValidationHandler();
			} else if (traceHandler.equals(Constants.TOTAL_TRACE_HANDLER)) {
				mySysErr.println("SETTING TOTAL TRACE HANDLER");
				traceObjectHandler = new TotalTraceHandler();
			} else if (traceHandler.equals(Constants.EXCLUDE_TRACE_HANDLER)) {
				mySysErr.println("SETTING CALLEE EXCLUDE TRACE HANDLER");
				traceObjectHandler = new ExcludeHandler(new ObjectTraceHandler());
			} else if (traceHandler
					.equals(Constants.CALLER_EXCLUDE_TRACE_HANDLER)) {
				mySysErr.println("SETTING CALLER EXCLUDE TRACE HANDLER");
				traceObjectHandler = new ExcludeHandler(
						new CallerObjectHandler());
			} else if (traceHandler
					.equals(Constants.CALLER_CLASS_EXCLUDE_TRACE_HANDLER)) {
				mySysErr.println("SETTING CALLER CLASS TRACE HANDLER");
				traceObjectHandler = new ExcludeHandler(
						new CallerClassHandler());
			} else if (traceHandler
					.equals(Constants.CALLEE_CLASS_EXCLUDE_TRACE_HANDLER)) {
				mySysErr.println("SETTING CALLEE CLASS TRACE HANDLER");
				traceObjectHandler = new ExcludeHandler(
						new CalleeClassHandler());
			} else if (traceHandler
					.equals(Constants.GLOBAL_WINDOW_EXCLUDE_TRACE_HANDLER)) {
				mySysErr.println("SETTING GLOBAL WINDOW TRACE HANDLER");
				traceObjectHandler = new ExcludeHandler(
						new GlobalWindowTraceHandler());
			} else if (traceHandler
					.equals(Constants.CALLE_HANDLER)) {
				mySysErr.println("SETTING CALLEE TRACE HANDLER");
				traceObjectHandler = new ObjectTraceHandler();
			}
		}

		resultFileLocation = System.getProperty(Constants.PROPERTY_RESULT_FILE);
		if (resultFileLocation == null) {
			resultFileLocation = DEFAULT_RESULT_FILE_LOCATION;
		}
		if (resultFileLocation.startsWith(Constants.COUNTER_PREFIX)) {
			resultFileLocation = getCounterFileName(resultFileLocation);

		}
		Runtime.getRuntime().addShutdownHook(new Thread() {
			public void run() {
				traceObjectHandler.report();
				mySysOut.println("SHUTDOWN\n" + traceCount + " calls traced");
			}
		});
	}

	private static String getCounterFileName(String filename) {
		filename = filename.replace(Constants.COUNTER_PREFIX, "");
		File file = new File(filename);
		File dir = file.getParentFile();
		final String filenamePrefix = file.getName().substring(0,
				file.getName().indexOf('.'));
		String[] filenames = dir.list(new FilenameFilter() {
			public boolean accept(File directory, String name) {
				if (name.startsWith(filenamePrefix)) {
					return true;
				}
				return false;
			}

		});
		mySysOut.println(Arrays.toString(filenames));
		int max = 0;
		for (int i = 0; i < filenames.length; i++) {
			if (getFileNumber(filenames[i], filenamePrefix) > max) {
				max = getFileNumber(filenames[i], filenamePrefix);
			}
		}
		return new File(dir, filenamePrefix + (max + 1) + ".ser").toString();
	}

	private static int getFileNumber(String name, String prefix) {
		try {
			String number = name.substring(name.indexOf(prefix)
					+ prefix.length(), name.lastIndexOf('.'));
			if (number.length() > 0) {
				return Integer.parseInt(number);
			}
		} catch (NumberFormatException e) {
			// ignore e.printStackTrace(mySysErr);
		}
		return 0;
	}

	public static void traceINSTR(int callerClassID, int callerMethodID,
			int callerObjectID, int calleeClassID, int calleeMethodID,
			int calleeObjectID) {
		traceObject = new TraceObject(callerClassID, callerMethodID,
				callerObjectID, calleeClassID, calleeMethodID, calleeObjectID);
		traceObjectHandler.handle(traceObject);
		traceCount++;
		if (0 == traceCount % 1e6) {
			mySysErr.println(traceCount + " CALLS");
		}
	}

	public static void traceFinalize() {
	}

	public static void objectArg(Object o) {
		mySysOut.println("OBJECT: " + o);
	}

	public static void intArg(int i) {
		mySysOut.println("INT: " + i);
	}

	public static class TraceObject implements Serializable {

		/**
		 *
		 */
		private static final long serialVersionUID = 1L;

		public int callerClassID;

		public int callerMethodID;

		public int callerObjectID;

		public int calleeClassID;

		public int calleeMethodID;

		public int calleeObjectID;

		public TraceObject(int callerClassID, int callerMethodID,
				int callerObjectID, int calleeClassID, int calleeMethodID,
				int calleeObjectID) {
			this.callerClassID = callerClassID;
			this.callerMethodID = callerMethodID;
			this.callerObjectID = callerObjectID;
			this.calleeClassID = calleeClassID;
			this.calleeMethodID = calleeMethodID;
			this.calleeObjectID = calleeObjectID;
		}

		public String toString() {
			StringBuffer buffer = new StringBuffer("\ncallerClassID: "
					+ callerClassID + "\ncallerMethodID: " + callerMethodID
					+ "\ncallerObjectID: ");
			if (callerObjectID == Constants.STATICOBJECTID) {
				buffer.append("STATIC");
			} else if (callerObjectID == Constants.CONSTRUCTORID) {
				buffer.append("CONSTRUCTOR: OBJECTID NOT INITIALIZED");
			} else {
				buffer.append(callerObjectID);
			}
			buffer.append("\ncalleeClassID: " + calleeClassID
					+ "\ncalleeMethodID: " + calleeMethodID
					+ "\ncalleeObjectID: ");
			if (calleeObjectID == Constants.STATICOBJECTID) {
				buffer.append("STATIC");
			} else if (calleeObjectID == Constants.CONSTRUCTORID) {
				buffer.append("CONSTRUCTOR: OBJECTID NOT INITIALIZED");
			} else {
				buffer.append(calleeObjectID);
			}
			return buffer.toString();

		}

		public boolean equals(Object obj) {
			if (obj instanceof TraceObject) {
				TraceObject compareTraceObject = (TraceObject) obj;
				return (compareTraceObject.callerClassID == callerClassID)
						|| (compareTraceObject.callerMethodID == callerMethodID)
						|| (compareTraceObject.callerObjectID == callerObjectID)
						|| (compareTraceObject.calleeClassID == calleeClassID)
						|| (compareTraceObject.calleeMethodID == calleeMethodID)
						|| (compareTraceObject.calleeObjectID == calleeObjectID);

			}
			return false;
		}
	}

	public static void insideAPI() {
		mySysOut.println("API TO API");

	}

	public void traceCall(int callerClassID, int callerMethodID,
			int callerObjectID, int calleeClassID, int calleeMethodID,
			int calleeObjectID) {
		traceINSTR(callerClassID, callerMethodID, callerObjectID,
				calleeClassID, calleeMethodID, calleeObjectID);
	}

	public void traceAPI() {
		insideAPI();
	}

	/**
	 * Indicates that tracer should start if this method is called the class is
	 * loaded in JVM and this starts the Tracer
	 */
	public static void start() {
		if (!started) {
			started = true;
			mySysOut.println("Tracing started");
		}

	}

	public static void println(Object o) {
		mySysOut.println(o);
	}
}
