package org.softevo.apibirthmark.tracing.mapping;

import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/**
 * IDToNameMap maps IDs to class and method names
 *
 * @author schuler
 *
 */
public class IDToNameMap {

	private Map<Integer, IDToNameEntry> idMap;

	/**
	 * Create a new IDToNameMap for given Map
	 *
	 * @param idMap
	 *            Map that contains the mapping of IDs to names
	 */
	public IDToNameMap(Map<Integer, IDToNameEntry> idMap) {
		this.idMap = idMap;
	}

	/**
	 * Return the name of the class with given class ID
	 *
	 * @param classID
	 *            ID of the class
	 * @return name of the class
	 */
	public String getClassnameByID(int classID) {
		IDToNameEntry entry = getIDToNameEntry(classID);
		return entry.className;
	}

	/**
	 * Return the name of the method for given class and method ID
	 *
	 * @param classID
	 *            ID of the class
	 * @param methodID
	 *            ID of the method
	 * @return name of the method
	 */
	public String getMethodnameByID(int classID, int methodID) {
		IDToNameEntry entry = getIDToNameEntry(classID);
		return entry.getMethodName(methodID);
	}

	/**
	 * Return the name of the method for given method ID
	 *
	 * @param methodID
	 * @return The name of the method
	 */
	public String getMethodnameByID(int methodID) {
		return getMethodnameByID(NameToIDEntry.getClassFromMethodID(methodID),
				methodID);

	}

	/**
	 * Return IDToNameEntry for given class ID
	 *
	 * @param classID
	 *            ID of the class
	 * @return the IDToNameEntry for the class
	 */
	private IDToNameEntry getIDToNameEntry(int classID) {
		if (idMap.containsKey(classID)) {
			return idMap.get(classID);
		}
		throw new IDMapError("Class ID not contained in hashMap\nclassID: "
				+ classID);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		Set keySet = idMap.keySet();
		Iterator it = keySet.iterator();
		StringBuffer buffer = new StringBuffer();
		while (it.hasNext()) {
			Integer i = (Integer) it.next();
			String className = (idMap.get(i)).className;
			buffer.append(i + "-" + className + "\n");
		}
		return buffer.toString();
	}

	/**
	 * Checks if given methodID is contained in IDToNameMap
	 *
	 * @param methodID
	 *            ID of the method
	 * @return
	 */
	public boolean contains(int methodID) {
		int classID = NameToIDEntry.getClassFromMethodID(methodID);
		Integer classKey = new Integer(classID);
		if (idMap.containsKey(classKey)) {
			IDToNameEntry classEntry = idMap.get(classKey);
			if (classEntry.methodHashMap.containsKey(new Integer(methodID))) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Checks if given methodID is contained in IDToNameMap
	 *
	 * @param classID
	 *            ID of the class
	 * @return
	 */
	public boolean containsClassID(int classID) {
		if (idMap.containsKey(classID)) {
			return true;
		}
		return false;
	}

	/**
	 * Add all entries of given IDToNameMap to this one. No duplicated entries
	 * are allowed
	 *
	 * @param idToNameMap
	 *            IDToNameMap to be added
	 */
	public void addMap(IDToNameMap idToNameMap) {
		for (Map.Entry<Integer, IDToNameEntry> element : idToNameMap.idMap
				.entrySet()) {
			if (idMap.containsKey(element.getKey())) {
				if (!idMap.get(element.getKey()).equals(element.getValue())) {
					throw new IDMapError("Element with key " + element.getKey()
							+ " already contained\n"
							+ idMap.get(element.getKey())
							+ "\n element to add" + element.getValue());
				}
			} else {
				idMap.put(element.getKey(), element.getValue());
			}
		}
	}

}
