package org.softevo.apibirthmark.tracing.mapping;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;


import com.thoughtworks.xstream.XStream;

import org.softevo.apibirthmark.constants.Constants;

/**
 * NameToIDMap does the mapping of class and method names to IDs
 *
 * @author schuler
 *
 */
public class NameToIDMap {

	private static Logger logger = Logger.getLogger(NameToIDMap.class);

	private HashMap<String, NameToIDEntry> classHashMap;

	private int maxClassKey;

	private int minClassKey;

	/**
	 * Create a new NameToIDMap
	 */
	public NameToIDMap() {
		int startKey = 129;
		classHashMap = new HashMap<String, NameToIDEntry>(200);
		NameToIDEntry noClassEntry = new NameToIDEntry(Constants.NO_METHOD_ID);
		classHashMap.put(Constants.NO_CLASSNAME, noClassEntry);
		insertMethod(Constants.NO_CLASSNAME, Constants.NO_METHODNAME, "!");
		maxClassKey = startKey;
		minClassKey = startKey;
	}

	/**
	 * Return the ClassMethodID for given class and method name.
	 *
	 * @param className
	 *            Name of the class
	 * @param methodName
	 *            Name of the method
	 * @param methodDescription
	 *            Method descriptor
	 * @return The ClassMethodID for class and method name
	 */
	public ClassMethodID getClassMethodID(String className, String methodName,
			String methodDescription) {
		int classID = getClassID(className);
		int methodID = getMethodID(className, methodName, methodDescription);
		return new ClassMethodID(classID, methodID);
	}

	/**
	 * Return method ID for given class and method name.
	 *
	 * @param className
	 *            Name of the class
	 * @param methodName
	 *            Name of the method
	 * @param methodDescription
	 *            Method descriptor
	 * @return The method ID for class and method name
	 */
	private int getMethodID(String className, String methodName,
			String methodDescription) {
		return getNameToIDEntry(className).getMethodID(methodName,
				methodDescription);
	}

	/**
	 * Return class ID for given class and method name.
	 *
	 * @param className
	 *            Name of the class
	 * @param methodName
	 *            Name of the method
	 * @param methodDescription
	 *            Method descriptor
	 * @return The class ID for class and method name
	 */

	public int getClassID(String className) {
		return getNameToIDEntry(className).getClassID();
	}

	/**
	 * Insert a new method into this NameToIDMap
	 *
	 * @param className
	 *            Name of the class
	 * @param methodName
	 *            Name of the method
	 * @param methodDescription
	 *            Method descriptor
	 */
	public void insertMethod(String className, String methodName,
			String methodDescription) {
		getNameToIDEntry(className).insertMethod(methodName, methodDescription);
	}

	/**
	 * Insert a new class into this NameToIDMap
	 *
	 * @param className
	 *            Name of the class
	 */
	public void insertClass(String className) {
		if (classHashMap.containsKey(className)) {
			throw new IDMapError(
					"Class already contained in NameToIDMap\nClass name: "
							+ className);
		}
		NameToIDEntry entry = new NameToIDEntry(maxClassKey);
		maxClassKey++;
		classHashMap.put(className, entry);
	}

	public NameToIDEntry getNameToIDEntry(String className) {
		if (classHashMap.containsKey(className)) {
			return classHashMap.get(className);
		}
		throw new IDMapError(
				"Class not contained in class HashMap\nClass name:  "
						+ className);
	}

	/**
	 * Insert a NameToIDEntry into this NameToIDMap for given class name.
	 *
	 * @param className
	 *            Name of the class
	 * @param nameToIDEntry
	 *            NameToIDEntry that is inserted
	 */
	public void insertNameToIDEntry(String className,
			NameToIDEntry nameToIDEntry) {
		classHashMap.put(className, nameToIDEntry);
	}

	/* ************** */
	/* FILE STUFF */
	/* ************** */

	/**
	 * Saves this NameToIDMap to file with given name
	 *
	 * @param filename
	 *            Name of the file to be saved
	 * @throws IOException
	 *             An IOException is thrown if something fails
	 */
	public void saveToFile(String filename) throws IOException {
		XStream xstream = new XStream();
		String xml = xstream.toXML(this);
		FileWriter filewriter = new FileWriter(new File(filename));
		BufferedWriter bufferedWriter = new BufferedWriter(filewriter);
		bufferedWriter.write(xml);
		bufferedWriter.close();
		filewriter.close();
	}

	/**
	 * Constructs NameToIDMap from given filename.
	 *
	 * @param filename
	 *            Name of the file
	 * @return The NameToIDMap for this file
	 * @throws IOException
	 *
	 */
	public static NameToIDMap getFromFile(String filename) throws IOException {
		FileReader fileReader = new FileReader(new File(filename));
		char[] charBuffer = new char[1024];
		StringBuffer xml = new StringBuffer();
		while (fileReader.read(charBuffer) != -1) {
			xml.append(charBuffer);
		}
		XStream xstream = new XStream();
		NameToIDMap idMap = (NameToIDMap) xstream.fromXML(xml.toString());
		fileReader.close();
		return idMap;
	}

	/**
	 * Return an IDToNameMap of this NameToIDMap
	 *
	 * @return The corresponding NameToIDMap
	 */
	public IDToNameMap getIDtoNameMap() {
		Set keySet = classHashMap.keySet();
		Iterator it = keySet.iterator();
		HashMap<Integer, IDToNameEntry> idHashMap = new HashMap<Integer, IDToNameEntry>();
		while (it.hasNext()) {
			String className = (String) it.next();
			NameToIDEntry element = classHashMap.get(className);

			HashMap<Integer, String> hm = element.getIDtoNameHashMap();
			idHashMap.put(new Integer(element.getClassID()), new IDToNameEntry(
					className, hm));
		}
		return new IDToNameMap(idHashMap);
	}

	/**
	 * Add all entries of the NameToIDMap contained in this File to this
	 * NameToIDMap
	 *
	 * @param filename
	 *            Name of the file where the NameToIDMap is stored
	 *
	 */
	public void appendFile(String filename) {
		try {
			NameToIDMap newIDMap = getFromFile(filename);
			addIDMap(newIDMap);
		} catch (IOException e) {
			e.printStackTrace();
		}

	}

	/**
	 * Add all entries of given NameToIDMap to this NameToIDMap
	 *
	 * @param newIDMap
	 *            NameToIDMap that contains the to be added Entries
	 */
	private void addIDMap(NameToIDMap newIDMap) {
		int minKey = newIDMap.getMinClassKey();
		if (minKey > maxClassKey) {
			HashMap<String, NameToIDEntry> addMap = newIDMap.getClassHashMap();
			classHashMap.putAll(addMap);
		}
	}

	/**
	 * Checks if class is contained in this NameToIDMap
	 *
	 * @param className
	 *            Name of the class
	 * @return true if the class is contained in this NameToIDMap
	 */
	public boolean contains(String className) {
		return classHashMap.containsKey(className);
	}

	/**
	 * Checks if method is contained in this NameToIDMap
	 *
	 * @param className
	 *            Name of the class
	 * @param methodName
	 *            Name of the method
	 * @param desc
	 *            Method Descriptor
	 * @return true if the method is contained in this NameToIDMap
	 */
	public boolean containsMethod(String className, String methodName,
			String desc) {
		if (contains(className)) {
			NameToIDEntry nameToIDEntry = classHashMap.get(className);
			return nameToIDEntry.containsMethod(methodName, desc);
		}
		return false;
	}

	/**
	 * Return the max class key. The highest ID that was used for a class
	 *
	 * @return The max class key
	 */
	public int getMaxClassKey() {
		return maxClassKey;
	}

	/**
	 * Set the max class key to specified number
	 *
	 * @param maxClassKey
	 *            the new max class key
	 */
	public void setMaxClassKey(int maxClassKey) {
		if (maxClassKey > 0 && maxClassKey >= this.maxClassKey) {
			this.maxClassKey = maxClassKey;
			this.minClassKey = maxClassKey;
		}
	}

	/**
	 * Return the lowest class ID used in this NameToIDMap.
	 *
	 * @return The lowest class ID
	 */
	public int getMinClassKey() {
		return minClassKey;
	}

	/**
	 * Return the Map that contains the mapping of names to NameToIDEntries.
	 *
	 * @return The Map that contains the mapping of names to NameToIDEntries
	 */
	public HashMap<String, NameToIDEntry> getClassHashMap() {
		return classHashMap;
	}

	/**
	 * Set the Map that contains he mapping of names to NameToIDEntries
	 *
	 * @param classHashMap
	 *            Map that contains the entries
	 */
	public void setClassHashMap(HashMap<String, NameToIDEntry> classHashMap) {
		this.classHashMap = classHashMap;
	}

	/**
	 * Combine two NameToIDMaps.
	 *
	 * @param firstMap
	 *            NameToIDMap to be combined
	 * @param secondMap
	 *            NameToIDMap to be combined
	 * @return One NameToIDMap that contains the entries of both maps
	 */
	public static NameToIDMap combine(NameToIDMap firstMap,
			NameToIDMap secondMap) {
		HashMap<String, NameToIDEntry> firstClassHashMap = firstMap
				.getClassHashMap();
		HashMap<String, NameToIDEntry> secondClassHashMap = secondMap
				.getClassHashMap();

		HashMap<String, NameToIDEntry> resultHashMap = new HashMap<String, NameToIDEntry>(
				firstClassHashMap);
		// resultHashMap.putAll(ht2);
		for (Map.Entry<String, NameToIDEntry> e : secondClassHashMap.entrySet()) {
			String key = e.getKey();
			if (firstClassHashMap.containsKey(key)) {
				if (((NameToIDEntry) firstClassHashMap.get(key)).getClassID() == ((NameToIDEntry) secondClassHashMap
						.get(key)).getClassID()) { // everything is fine but
					// don't add to result
					continue;
				}
				while (firstClassHashMap.containsKey(key)) {
					logger.warn(" Double Entry: " + key);
					key = key + " Double Entry";
				}
			}
			resultHashMap.put(key, e.getValue());
		}
		NameToIDMap resultIDMap = new NameToIDMap();
		resultIDMap.setMaxClassKey(Math.max(secondMap.getMaxClassKey(),
				firstMap.getMaxClassKey()));
		resultIDMap.setClassHashMap(resultHashMap);
		resultIDMap.isSane();
		return resultIDMap;
	}

	/**
	 * Checks if this NameToIDMap is in a sane state. If not an RuntimeException
	 * is thrown.
	 */
	private void isSane() {
		Set<String> classNames = new HashSet<String>();
		Map<Integer, String> ids = new HashMap<Integer, String>();
		for (Map.Entry<String, NameToIDEntry> element : classHashMap.entrySet()) {
			String key = element.getKey();
			Integer value = new Integer(element.getValue().getClassID());
			if (classNames.contains(key)) {
				throw new IDMapError("class contained 2 times" + "\n key: "
						+ key + "\n value: " + value);
			} else {
				classNames.add(key);
			}
			if (ids.containsKey(value)) {
				NameToIDEntry nameToIDEntry = (NameToIDEntry) classHashMap
						.get(key);
				throw new IDMapError("same ID for 2 classes" + "\n key: " + key
						+ "\n value: " + value + " (duplicate to)"
						+ ids.get(value) + "\n" + nameToIDEntry.getClassID());
			} else {
				ids.put(value, key);
			}
		}
	}

	/**
	 * Read a NameToIDMap from a file and return the corresponding IDToNameMap
	 * of it
	 *
	 * @param filename
	 *            Name of the file that contains the NameToIDMap
	 * @return The corresponding IDToNameMap
	 */
	public static IDToNameMap getIDToNameMapfromNameToIDFile(String filename) {
		IDToNameMap idToNameMap = null;
		try {
			idToNameMap = (NameToIDMap.getFromFile(filename)).getIDtoNameMap();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return idToNameMap;
	}

	/**
	 * Inserts the class into this NameToIDMap, if not already contained.
	 *
	 * Return the ClassMethodID for the class and method.
	 *
	 * @param className
	 *            Name of the class
	 * @param methodName
	 *            Name of the method
	 * @param desc
	 *            Method Descriptor
	 * @return The ClassMethodID for this class and method
	 */
	public ClassMethodID getAndInsertClassMethodID(String className,
			String methodName, String desc) {
		if (containsMethod(className, methodName, desc)) {
			return getClassMethodID(className, methodName, desc);
		} else {
			if (!contains(className)) {
				insertClass(className);
			}
			insertMethod(className, methodName, desc);
			return getClassMethodID(className, methodName, desc);
		}
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuffer buffer = new StringBuffer();
		for (Map.Entry<String, NameToIDEntry> element : classHashMap.entrySet()) {
			buffer.append("\n" + element.getKey());
		}
		return buffer.toString();
	}
}
