package org.softevo.apibirthmark.tracing.result;

import java.io.Serializable;

import org.softevo.apibirthmark.tracing.mapping.IDToNameMap;
import org.softevo.apibirthmark.tracing.tracehandler.window.AbstractResultWindow;

/**
 * An AbstractTraceResult holds the results for tracing a program run. It
 * contains all call sequences. Thus it represents the call sequence set.
 *
 * @author schuler
 *
 */
public abstract class AbstractTraceResult implements Serializable {
	private static final long serialVersionUID = 1;

	/**
	 * Return a copy of this AbstractTraceResult
	 *
	 * @return
	 */
	public abstract AbstractTraceResult copy();

	/**
	 * Return a String representing this AbstractTraceResult. Where all
	 * methoodIDs that are contained in given IDToNameMap are replaced with
	 * their actual names.
	 *
	 * @param idToNameMap
	 *            {@link IDToNameMap} that contains the mapping of IDs to names
	 * @return
	 */
	public abstract String toNameString(IDToNameMap idToNameMap);

	/**
	 * Add all sequences of given AbstractTraceResult to this one.
	 *
	 * @param addResult
	 *            The AbstractTraceResult to add
	 */
	public abstract void add(AbstractTraceResult addResult);

	/**
	 * Subtract all sequences of given AbstractTraceResult from this one.
	 *
	 * @param subtractResult
	 *            The AbstractTraceResult to subtract
	 */
	public abstract void subtract(AbstractTraceResult subtractResult);

	/**
	 * Intersect this AbstractTraceResult with the one that is given
	 *
	 * @param intersectResult
	 *            The AbstractTraceResult to intersect with
	 */
	public abstract void intersect(AbstractTraceResult intersectResult);

	/**
	 * Return the number of sequences contained in this set.(not the number of
	 * different callee sequences)
	 *
	 * @return The number of sequences
	 */
	public abstract int getNumberOfSequences();

	/**
	 * Return the sequences of this set
	 *
	 * @return The call sequences
	 */
	public abstract AbstractResultWindow[] getSequences();

	/**
	 * Find a sequence in this set
	 *
	 * @param findSequence
	 *            The sequence to find
	 * @return Position of the sequence in the array, if it is not contained
	 *         return -1.
	 */
	public abstract int find(AbstractResultWindow findSequence);

	/**
	 * Add two AbstractTraceResults and return the result
	 *
	 * @param firstResult
	 *            AbstractTraceResult to add
	 * @param secondResult
	 *            AbstractTraceResultto add
	 * @return Result of the addition
	 */
	public static AbstractTraceResult add(AbstractTraceResult firstResult,
			AbstractTraceResult secondResult) {
		AbstractTraceResult copy = firstResult.copy();
		copy.add(secondResult);
		return copy;
	}

	/**
	 * Subtract two AbstractTraceResults and return the result
	 *
	 * @param firstResult
	 *            AbstractTraceResult to add
	 * @param secondResult
	 *            AbstractTraceResultto add
	 * @return Result of the subtraction
	 */
	public static AbstractTraceResult subtract(AbstractTraceResult firstResult,
			AbstractTraceResult secondResult) {
		AbstractTraceResult copy = firstResult.copy();
		copy.subtract(secondResult);
		return copy;
	}

	/**
	 * Intersect two AbstractTraceResults and return the result
	 *
	 * @param firstResult
	 *            AbstractTraceResult to add
	 * @param secondResult
	 *            AbstractTraceResultto add
	 * @return Result of the intersection
	 */
	public static AbstractTraceResult intersect(
			AbstractTraceResult firstResult, AbstractTraceResult secondResult) {
		AbstractTraceResult copy = firstResult.copy();
		copy.intersect(secondResult);
		return copy;
	}

	/**
	 * Return the number of differing callee sequences.
	 *
	 * @return The number of differing callee sequences.
	 */
	public abstract int getNumberOfCalleeSequences();

	/**
	 * Return a new TraceObjectResult with all sequences in the firstResult that
	 * have a matching callee sequences in the second Result. Note:
	 * intersectByCalleeID(X,Y) will differ from intersectByCalleeID(Y,X) in
	 * sequences and also in the number of sequences.
	 *
	 * @param firstResult
	 * @param secondResult
	 * @return
	 */
	public static AbstractTraceResult intersectByCalleeID(
			AbstractTraceResult firstResult, AbstractTraceResult secondResult) {
		AbstractTraceResult copy = firstResult.copy();
		copy.intersectByCalleeID(secondResult);
		return copy;

	}

	/**
	 * All sequences that do not have a matching callee-sequence in
	 * intersectResult are removed.
	 *
	 * @param intersectResult
	 *            Result that is intersected with.
	 */
	public abstract void intersectByCalleeID(AbstractTraceResult intersectResult);

	/**
	 * Return the window size of the sequences
	 *
	 * @return The seqeunce's window size
	 */
	public abstract int getWindowSize();

	/**
	 * Return a String with names or IDs( if ID not contained in IDToNameMap) of
	 * all classes that are contained in sequences
	 *
	 * @param idMap
	 * @return String of contained class names and IDs
	 */
	public abstract String getClasses(IDToNameMap idMap);
}