package org.softevo.apibirthmark.tracing.result;

import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.Vector;
import java.util.Map.Entry;

import org.softevo.apibirthmark.tracing.mapping.IDToNameMap;
import org.softevo.apibirthmark.tracing.mapping.NameToIDEntry;
import org.softevo.apibirthmark.tracing.tracehandler.window.AbstractResultWindow;
import org.softevo.apibirthmark.tracing.tracehandler.window.TraceObjectResultWindow;

/**
 * TraceObjectResult holds the results for tracing a program run. It contains
 * all call sequences. Thus it represents the call sequence set
 *
 * @author schuler
 *
 */
public class TraceObjectResult extends AbstractTraceResult implements Cloneable {

	private static final long serialVersionUID = 1;

	/**
	 * Array that contains the call sequences
	 */
	private AbstractResultWindow[] sequences;

	/**
	 * Indicates if sequences are sorted
	 */
	private boolean sorted = false;

	/**
	 * Create an empty TraceObjectResult
	 */
	public TraceObjectResult() {
		sequences = new TraceObjectResultWindow[0];
	}

	/**
	 * Create new TraceObjectResultWindow for given sequences
	 *
	 * @param resultWindow
	 *            Sequences that were observed
	 */
	public TraceObjectResult(AbstractResultWindow[] resultWindow) {
		sequences = resultWindow;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see tracing.result.AbstractTraceResult#toNameString(tracing.mapping.IDToNameMap)
	 */
	@Override
	public String toNameString(IDToNameMap idToNameMap) {
		StringBuilder buffer = new StringBuilder("\n");
		for (AbstractResultWindow abstractTraceResult : sequences) {
			buffer.append(abstractTraceResult.toNameString(idToNameMap));
			buffer.append('\n');
			buffer.append('\n');
		}
		return buffer.toString();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see tracing.result.AbstractTraceResult#subtract(tracing.result.AbstractTraceResult)
	 */
	@Override
	public void subtract(AbstractTraceResult subtractResult) {
		Vector<AbstractResultWindow> subtractVector = new Vector<AbstractResultWindow>();
		for (AbstractResultWindow subtractSequence : sequences) {
			if (subtractResult.find(subtractSequence) < 0) {
				subtractVector.add(subtractSequence);
			}

		}
		sequences = subtractVector.toArray(new AbstractResultWindow[0]);
		sorted = false;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see tracing.result.AbstractTraceResult#intersect(tracing.result.AbstractTraceResult)
	 */
	@Override
	public void intersect(AbstractTraceResult intersectResult) {
		Vector<AbstractResultWindow> intersectVector = new Vector<AbstractResultWindow>();
		for (AbstractResultWindow intersectSequence : intersectResult
				.getSequences()) {
			if (find(intersectSequence) >= 0) {
				intersectVector.add(intersectSequence);
			}
		}
		sequences = intersectVector.toArray(new AbstractResultWindow[0]);
		sorted = false;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see tracing.result.AbstractTraceResult#find(tracing.tracehandler.window.AbstractResultWindow)
	 */
	@Override
	public int find(AbstractResultWindow intersectSequence) {
		if (!sorted) {
			sort();
		}
		return Arrays.binarySearch(sequences, intersectSequence);
	}

	/**
	 * Sort the array that contains the sequences
	 */
	private void sort() {
		Arrays.sort(sequences);
		sorted = true;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see tracing.result.AbstractTraceResult#getNumberOfSequences()
	 */
	@Override
	public int getNumberOfSequences() {
		return sequences.length;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see tracing.result.AbstractTraceResult#getSequences()
	 */
	@Override
	public AbstractResultWindow[] getSequences() {
		return sequences;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see tracing.result.AbstractTraceResult#add(tracing.result.AbstractTraceResult)
	 */
	@Override
	public void add(AbstractTraceResult addResult) {
		Vector<AbstractResultWindow> addVector = new Vector<AbstractResultWindow>();
		for (AbstractResultWindow addSequence : addResult.getSequences()) {
			if (find(addSequence) < 0) {
				addVector.add(addSequence);
			}
		}
		addVector.addAll(Arrays.asList(sequences));
		sequences = addVector.toArray(new AbstractResultWindow[0]);
		sorted = false;
	}

	public TraceObjectResult[] getSameCalleeSequences() {
		HashMap<CallSequence, Vector<AbstractResultWindow>> calleeSequenceMap = new HashMap<CallSequence, Vector<AbstractResultWindow>>();
		for (AbstractResultWindow resultWindow : sequences) {
			CallSequence calleeSequence = new CallSequence(resultWindow
					.getCalleeSequence());
			if (calleeSequenceMap.containsKey(calleeSequence)) {
				calleeSequenceMap.get(calleeSequence).add(resultWindow);
			} else {
				Vector<AbstractResultWindow> vector = new Vector<AbstractResultWindow>();
				vector.add(resultWindow);
				calleeSequenceMap.put(calleeSequence, vector);
			}
		}
		Set<Entry<CallSequence, Vector<AbstractResultWindow>>> s = calleeSequenceMap
				.entrySet();
		TraceObjectResult[] results = new TraceObjectResult[s.size()];
		int pos = 0;
		for (Entry<CallSequence, Vector<AbstractResultWindow>> entry : s) {
			AbstractResultWindow[] array = entry.getValue().toArray(
					new AbstractResultWindow[0]);
			results[pos] = new TraceObjectResult(array);
			pos++;
		}
		return results;
	}

	public void filterByCaller(IDToNameMap callerMap) {
		Vector<AbstractResultWindow> resultVector = new Vector<AbstractResultWindow>();
		for (AbstractResultWindow resultWindow : sequences) {
			int[] callerSequence = resultWindow.getCallerClassSequence();
			for (int i : callerSequence) {
				if (callerMap.contains(i)) {
					resultVector.add(resultWindow);
					break;
				}
			}
		}
		sequences = resultVector.toArray(new AbstractResultWindow[0]);
		sorted = false;
	}

	/**
	 * CallSequence can hold one call sequence (int array) so that sequences can
	 * be compared
	 *
	 * @author schuler
	 *
	 */
	private static class CallSequence {
		int[] sequence;

		public CallSequence(int[] array) {
			sequence = array;
		}

		/*
		 * (non-Javadoc)
		 *
		 * @see java.lang.Object#equals(java.lang.Object)
		 */
		@Override
		public boolean equals(Object obj) {
			CallSequence compareCalleeSequence = (CallSequence) obj;
			return Arrays.equals(sequence, compareCalleeSequence.sequence);

		}

		/*
		 * (non-Javadoc)
		 *
		 * @see java.lang.Object#hashCode()
		 */
		@Override
		public int hashCode() {
			int hash = 0;
			for (int i : sequence) {
				hash += i;
			}
			return hash;
		}

	}

	/*
	 * (non-Javadoc)
	 *
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder buffer = new StringBuilder();
		for (AbstractResultWindow resultWindow : sequences) {
			buffer.append(resultWindow.toString());
			buffer.append('\n');
		}
		return buffer.toString();
	}

	/**
	 * Return a copy of this TraceObjectResult
	 *
	 */
	@Override
	public TraceObjectResult copy() {
		TraceObjectResultWindow[] newSequences = new TraceObjectResultWindow[sequences.length];
		System.arraycopy(sequences, 0, newSequences, 0, newSequences.length);
		return new TraceObjectResult(newSequences);
	}

	/**
	 * All sequences that do not have a matching callee-sequence in
	 * intersectResult are removed.
	 *
	 * @param intersectResult Result that is intersected with.
	 */
	@Override
	public void intersectByCalleeID(AbstractTraceResult intersectResult) {
		if (sequences.length == 0) {
			// logger.error("result contains no sequences");
			return;
		}
		if (intersectResult.getSequences().length == 0) {
			// logger.error("intersect result contains no sequences");
			sequences = intersectResult.getSequences();
		}

		if (getWindowSize() != intersectResult.getWindowSize()) {
			throw new RuntimeException("Window sizes do not match:"
					+ getWindowSize() + " vs. "
					+ intersectResult.getWindowSize());
		}
		HashSet<CallSequence> intersectCalleeSequences = new HashSet<CallSequence>();
		for (AbstractResultWindow resultWindow : intersectResult.getSequences()) {
			CallSequence calleeSequence = new CallSequence(resultWindow
					.getCalleeSequence());
			if (!intersectCalleeSequences.contains(calleeSequence)) {
				intersectCalleeSequences.add(calleeSequence);
			}
		}
		Vector<AbstractResultWindow> resultVector = new Vector<AbstractResultWindow>();
		for (AbstractResultWindow resultWindow : sequences) {
			CallSequence calleeSequence = new CallSequence(resultWindow
					.getCalleeSequence());
			if (intersectCalleeSequences.contains(calleeSequence)) {
				resultVector.add(resultWindow);
			}
		}
		sequences = resultVector.toArray(new AbstractResultWindow[0]);
		sorted = false;
	}

	/**
	 * Remove all sequences that have an equal callee sequence in subtractResult
	 *
	 * @param subtractResult
	 */
	public void subtractByCalleeID(TraceObjectResult subtractResult) {
		HashSet<CallSequence> subtractSequences = new HashSet<CallSequence>();
		for (AbstractResultWindow window : subtractResult.sequences) {
			subtractSequences.add(new CallSequence(window.getCalleeSequence()));
		}

		Vector<AbstractResultWindow> remainingVector = new Vector<AbstractResultWindow>();
		for (AbstractResultWindow resultWindow : sequences) {
			CallSequence callSequence = new CallSequence(resultWindow
					.getCalleeSequence());
			if (!subtractSequences.contains(callSequence)) {
				remainingVector.add(resultWindow);
			}

		}
		sequences = remainingVector.toArray(new AbstractResultWindow[0]);
		sorted = false;
	}





	/**
	 * Return the number of sequences that contain given classID
	 *
	 * @param classID
	 * @return
	 */
	public int getNumberOfClassSequences(int classID) {
		int result = 0;
		for (AbstractResultWindow window : sequences) {
			if (window.containsClassID(classID)) {
				result++;
			}
		}
		return result;
	}

	/**
	 * Return all sequences that contain given classID
	 *
	 * @param classID
	 *            ID of the class
	 * @return TraceObjectResult that contains all sequences that contain given
	 *         class
	 */
	public TraceObjectResult getClassSequences(int classID) {
		Vector<AbstractResultWindow> vector = new Vector<AbstractResultWindow>();
		for (AbstractResultWindow window : sequences) {
			if (window.containsClassID(classID)) {
				vector.add(window);
			}
		}
		return new TraceObjectResult(vector
				.toArray(new AbstractResultWindow[0]));
	}

	/**
	 * Return a String with names or IDs( if ID not contained in IDToNameMap) of
	 * all classes that are contained in sequences
	 *
	 * @param idMap
	 * @return String of contained class names and IDs
	 */
	public String getClasses(IDToNameMap idMap) {
		StringBuilder buffer = new StringBuilder();
		Set<Integer> callerSet = new HashSet<Integer>();
		Set<Integer> calleeSet = new HashSet<Integer>();
		for (AbstractResultWindow window : sequences) {
			for (int i : window.getCallerMethodSequence()) {
				callerSet.add(NameToIDEntry.getClassFromMethodID(i));
			}
		}
		for (AbstractResultWindow window : sequences) {
			for (int i : window.getCalleeSequence()) {
				calleeSet.add(NameToIDEntry.getClassFromMethodID(i));
			}
		}
		buffer.append("Caller Classes\n");
		for (Integer integer : callerSet) {
			if (idMap.containsClassID(integer)) {
				buffer.append(idMap.getClassnameByID(integer)).append('\n');
			}
		}
		buffer.append("Callee Classes\n");
		for (Integer integer : calleeSet) {
			if (idMap.containsClassID(integer)) {
				buffer.append(idMap.getClassnameByID(integer)).append('\n');
			}
		}
		return buffer.toString();
	}

	/**
	 * Return the number of differing callee sequences.
	 *
	 * @return  The number of differing callee sequences.
	 */
	@Override
	public int getNumberOfCalleeSequences() {
		Set<CallSequence> sequenceSet = new HashSet<CallSequence>();
		for (AbstractResultWindow sequence : sequences) {
			sequenceSet.add(new CallSequence(sequence.getCalleeSequence()));
		}
		return sequenceSet.size();
	}

	/**
	 * Return a TraceObjectResult where each callee sequence of this
	 * TraceObjectResult is contained exactly once.
	 *
	 * @return The TraceObjectResult that contains every callee sequence once
	 */
	public TraceObjectResult getCaleeSequences() {
		Set<CallSequence> sequenceSet = new HashSet<CallSequence>();
		Vector<AbstractResultWindow> resultWindowVector = new Vector<AbstractResultWindow>();
		for (AbstractResultWindow sequence : sequences) {
			if (!sequenceSet.contains(new CallSequence(sequence
					.getCalleeSequence()))) {
				resultWindowVector.add(sequence);
			}
			sequenceSet.add(new CallSequence(sequence.getCalleeSequence()));
		}
		TraceObjectResult calleeResult = new TraceObjectResult(
				resultWindowVector.toArray(new AbstractResultWindow[0]));
		return calleeResult;
	}

	/**
	 * subtract all callee sequences that are contained in firstResult and
	 * secondResult from firstResult
	 *
	 * @param firstResult
	 * @param secondResult
	 * @return a new TraceObjectResult = r1 -r2
	 */
	public static TraceObjectResult subtractByCalleeID(
			TraceObjectResult firstResult, TraceObjectResult secondResult) {
		TraceObjectResult firstResultCopy = firstResult.copy();
		firstResultCopy.subtractByCalleeID(secondResult);
		return firstResultCopy;
	}

	/**
	 * Return the window size of the sequences
	 *
	 * @return The seqeunce's window size
	 */
	@Override
	public int getWindowSize() {
		if (sequences != null && sequences.length > 0) {
			return sequences[0].getSize();
		}
		return 0;
	}

	/**
	 * Return all caller classIDs
	 *
	 * @return
	 */
	public int[] getCallerClasses() {
		Set<Integer> callerClassSet = new HashSet<Integer>();
		for (AbstractResultWindow resultWindow : sequences) {
			int[] classes = resultWindow.getCallerClassSequence();
			for (int i : classes) {
				if (!callerClassSet.contains(i)) {
					callerClassSet.add(i);
				}
			}
		}
		int[] retArray = new int[callerClassSet.size()];
		int index = 0;
		for (Integer integer : callerClassSet) {
			retArray[index] = integer;
			index++;
		}
		return retArray;
	}

	/**
	 * Return names of caller classes if class id is contained in given
	 * NameToIDMap
	 *
	 * @param toNameMap
	 * @return
	 */
	/**
	 * @param toNameMap
	 * @return
	 */
	public String getCallerClasses(IDToNameMap toNameMap) {
		int[] callerClassIDs = getCallerClasses();
		StringBuilder buffer = new StringBuilder();
		Vector<String> classes = new Vector<String>();
		for (int id : callerClassIDs) {
			StringBuilder classBuffer = new StringBuilder();
			if (toNameMap.containsClassID(id)) {
				classBuffer.append(toNameMap.getClassnameByID(id));
			} else {
				classBuffer.append(id + "");
			}
			classBuffer.append("  "
					+ getClassSequences(id).getNumberOfCalleeSequences()
					+ "  id: " + id);
			classes.add(classBuffer.toString());
		}
		String[] classesArray = classes.toArray(new String[0]);
		Arrays.sort(classesArray);
		for (String classname : classesArray) {
			buffer.append(classname).append('\n');
		}
		return buffer.toString();
	}

}
