package org.softevo.apibirthmark.tracing.result.process;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Collection;
import java.util.Formatter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import org.softevo.apibirthmark.tracing.result.AbstractTraceResult;
import org.softevo.apibirthmark.tracing.result.birthmark.Birthmark;
import org.softevo.apibirthmark.tracing.result.birthmark.callee.CalleeSequenceBirthmark;
import org.softevo.apibirthmark.tracing.result.birthmark.callee.ContainedCalleeSequenceBirthmark;
import org.softevo.apibirthmark.util.ObjectIO;

/**
 * Compute the Birthmarks for given results and a compare result
 *
 * @author schuler
 *
 */
public class ComputeBirthmarks {

	private static final String TABLE_FLOAT_FORMAT = "  %.5f  :";

	private static final String FLOAT_FORMAT = "%.5f";

	private static final int NAME_LENGTH = 10;

	private static final String NAME_FORMAT = "  %-" + NAME_LENGTH + "s";

	/**
	 * used to control the birthmark similarity, if birthmark_sim(a,b) =
	 * birthmark_sim(b,a). Note that this is not the case for some similarity
	 * measures
	 */
	private static boolean controllBirthmarks = false;

	/**
	 * Compute the given Birthmark between compareResult and all
	 * trainingResults.
	 *
	 * @param birthmark
	 * @param trainingResults
	 * @param compareResult
	 */
	public static Map<String, Float> computeBirthmark(Birthmark birthmark,
			Map<String, AbstractTraceResult> trainingResults,
			AbstractTraceResult compareResult) {
		float[] birthmarks = new float[trainingResults.size()];
		Formatter floatFormater = new Formatter(new StringBuilder());
		Formatter controlFormater = new Formatter(new StringBuilder());
		Formatter resultNames = new Formatter(new StringBuilder());
		Map<String, Float> resultMap = new HashMap<String, Float>();
		int index = 0;
		birthmark.setCompareResult(compareResult);
		for (Map.Entry<String, AbstractTraceResult> entry : trainingResults
				.entrySet()) {
			birthmark.setFirstResult(entry.getValue());
			birthmark.computeBirthmark();
			float birthmarkValue = birthmark.getBirthmarkValue();
			if (controllBirthmarks) {
				birthmark.setFirstResult(compareResult);
				birthmark.setCompareResult(entry.getValue());
				birthmark.computeBirthmark();
				float controlValue = birthmark.getBirthmarkValue();
				controlFormater.format(FLOAT_FORMAT, controlValue);
				birthmark.setCompareResult(compareResult);
				if (controlValue != birthmarkValue) {
					throw new RuntimeException("Birthmarks unequal"
							+ birthmarkValue + " " + controlValue);
				}
			}
			birthmarks[index++] = birthmarkValue;
			resultNames.format(NAME_FORMAT, stripLength(entry.getKey(),
					NAME_LENGTH));
			floatFormater.format(TABLE_FLOAT_FORMAT, birthmarkValue);
			System.out.printf(("%-30s  " + FLOAT_FORMAT + "\n"),
					entry.getKey(), birthmarkValue);
			resultMap.put(entry.getKey(), birthmarkValue);
		}
		floatFormater.format("\n");

		System.out.println(resultNames);
		System.out.println(floatFormater.toString());
		return resultMap;
	}



	/**
	 * Compute all Birthmarks between compareResult and all trainingResults.
	 *
	 * @param trainingResults
	 * @param compareResult
	 */
	public static void computeAllBirthmarks(
			Map<String, AbstractTraceResult> trainingResults,
			AbstractTraceResult compareResult) {
		Collection<AbstractTraceResult> resultSet = new HashSet<AbstractTraceResult>(
				trainingResults.values());
		resultSet.remove(compareResult);
		Birthmark[] birthmarks = new Birthmark[] {
				new CalleeSequenceBirthmark(compareResult, null),
				new ContainedCalleeSequenceBirthmark(null, null), };
		for (Birthmark birthmark : birthmarks) {
			String classname = birthmark.getClass().getCanonicalName();
			System.out.println(getClassName(classname));
			computeBirthmark(birthmark, trainingResults, compareResult);
		}
	}

	/**
	 * Return the suffix (of given length) of a string if the length of the
	 * string is bigger than the given length else just return the string
	 *
	 * @param string
	 * @param length
	 * @return
	 */
	private static String stripLength(String string, int length) {
		if (string.length() > length) {
			return string.substring(string.length() - length);
		}
		return string;
	}

	/**
	 * Return just the last part of a full qualified class name. foo.bar.class ->
	 * class
	 *
	 * @param classname
	 * @return
	 */
	private static String getClassName(String classname) {
		return classname.substring(classname.lastIndexOf('.') + 1);
	}

	/**
	 * Compute the Birthmark between two results (read from given filenames)
	 *
	 * @param birthmark
	 * @param firstResultFilename
	 * @param compareResultFilename
	 */
	public static float computeBirthmark(Birthmark birthmark,
			String firstResultFilename, String compareResultFilename) {
		try {
			AbstractTraceResult firstTraceObjectResult = (AbstractTraceResult) ObjectIO
					.getFromFile(firstResultFilename);
			AbstractTraceResult secondObjectResult = (AbstractTraceResult) ObjectIO
					.getFromFile(compareResultFilename);
			birthmark.setFirstResult(firstTraceObjectResult);
			birthmark.setCompareResult(secondObjectResult);
			birthmark.computeBirthmark();
			float birthmarkValue = birthmark.getBirthmarkValue();
			System.out.println("Birthmark value " + birthmarkValue);
			System.out.println("files compared:\n" + firstResultFilename + "\n"
					+ compareResultFilename);
			return birthmarkValue;
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
		return 0;

	}

	/**
	 * Compute the Birthmark between two trace results
	 *
	 * @param originalResult
	 * @param obfuscatedResult
	 * @return
	 */
	public static float computeBirtmark(AbstractTraceResult originalResult,
			AbstractTraceResult obfuscatedResult) {
		if (originalResult == null || obfuscatedResult == null) {
			return -1;
		}
		Birthmark birthmark = new CalleeSequenceBirthmark(originalResult,
				obfuscatedResult);
		birthmark.computeBirthmark();
		return birthmark.getBirthmarkValue();
	}

}
