package org.softevo.apibirthmark.tracing.result.process;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FilenameFilter;
import java.io.IOException;
import java.sql.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.apache.tools.ant.DirectoryScanner;

import org.softevo.apibirthmark.tracing.mapping.IDToNameMap;
import org.softevo.apibirthmark.tracing.mapping.NameToIDMap;
import org.softevo.apibirthmark.tracing.result.AbstractTraceResult;
import org.softevo.apibirthmark.util.ObjectIO;
import org.softevo.apibirthmark.util.Property;

/**
 * Class to get trace results from files and compute birthmarks between them
 *
 * @author schuler
 *
 */
public class ResultFromFiles {

	private static Logger logger = Logger.getLogger(ResultFromFiles.class);

	private static Map<File, AbstractTraceResult> resultCache = new HashMap<File, AbstractTraceResult>();

	/**
	 * show class names for every TraceObjectResult in given map
	 *
	 * @param idFileName
	 * @param map
	 */
	public static void showClassNames(String idFileName,
			Map<String, AbstractTraceResult> map) {
		IDToNameMap idMap = NameToIDMap
				.getIDToNameMapfromNameToIDFile(idFileName);
		for (Map.Entry<String, AbstractTraceResult> entry : map.entrySet()) {
			System.out.println(entry.getKey());
			String classes = entry.getValue().getClasses(idMap);
			System.out.println("CLASSES:\n" + classes);
		}
	}

	/**
	 * Computes all birthmarks between map and compare result and optionally
	 * removes compareResult from map.
	 *
	 * @param map
	 * @param compareResultKey
	 * @param removeCompareResult
	 */
	public static void computeAllBirthmarks(
			Map<String, AbstractTraceResult> map, String compareResultKey,
			boolean removeCompareResult) {
		AbstractTraceResult compareResult;
		if (removeCompareResult) {
			compareResult = map.remove(compareResultKey);
		} else {
			compareResult = map.get(compareResultKey);
		}
		ComputeBirthmarks.computeAllBirthmarks(map, compareResult);
	}

	/**
	 * Get results recursively form a directory
	 *
	 * @param directory
	 * @return
	 */
	public static Map<String, AbstractTraceResult> getResultsFromDirectory(
			File directory) {
		if (directory.isDirectory()) {
			System.out.println("Reading from " + directory);
		}
		Map<String, File> fileMap = getFileMap(getFilesFromDirectoryRecursively(directory));
		return getTraceObjectResultMap(fileMap);
	}

	public static Map<String, AbstractTraceResult> getResultsFromDirectoryNonRecursive(
			File directory) {
		Map<String, File> fileMap = getFilesFromDirectory(directory);
		return getTraceObjectResultMap(fileMap);
	}

	/**
	 * Return a Map with same keys and TraceObjectResults (read from given
	 * Files)
	 *
	 * @param fileMap
	 * @return
	 */
	private static Map<String, AbstractTraceResult> getTraceObjectResultMap(
			Map<String, File> fileMap) {
		Map<String, AbstractTraceResult> resultMap = new HashMap<String, AbstractTraceResult>();
		for (Map.Entry<String, File> entry : fileMap.entrySet()) {
			resultMap.put(entry.getKey(),
					getTraceObjectResult(entry.getValue()));
		}
		return resultMap;
	}

	/**
	 * Get recursively all files in a given directory an return them in a Map
	 * (with filenames as keys)
	 *
	 * @param directory
	 * @return
	 */
	private static Map<String, File> getFileMap(File[] files) {
		Map<String, File> filemap = new HashMap<String, File>();
		for (File resultFile : files) {
			filemap.put(removeClutter(resultFile.getName()), resultFile);
		}
		return filemap;
	}

	/**
	 * Return all result files starting at given directory.The directory is
	 * recursively scanned for all result files
	 *
	 * @param dir
	 * @return
	 */
	private static File[] getFilesFromDirectoryRecursively(File dir) {
		DirectoryScanner ds = new DirectoryScanner();
		String[] includes = new String[] { "**/*.ser" };
		ds.setIncludes(includes);
		ds.setBasedir(dir);
		ds.setCaseSensitive(true);
		ds.scan();
		String[] filenames = ds.getIncludedFiles();
		Vector<File> files = new Vector<File>();
		for (String filename : filenames) {
			files.add(new File(dir, filename));
		}
		return files.toArray(new File[0]);
	}

	/**
	 * Return all results from the property file
	 *
	 * @return
	 */
	private static Map<String, File> getResults() {
		String[] keys = Property.getKeys("result");
		Map<String, File> fileMap = new HashMap<String, File>();
		for (String key : keys) {
			String filename = Property.getProperty(key);
			File file = new File(filename);
			if (file.exists()) {
				if (file.isDirectory()) {
					fileMap.putAll(getFilesFromDirectory(file));
				} else {
					fileMap.put(removeClutter(key), file);
				}
			} else {
			}
		}
		removeDuplicates(fileMap);
		return fileMap;
	}

	/**
	 * Remove unnecessary parts from a filename so that it can serve as a
	 * descriptive key
	 *
	 * @param filename
	 * @return
	 */
	public static String removeClutter(String filename) {
		String[] clutter = new String[] { ".ser", "pngReader_", "pngReader",
				".result", "ImageReader_", "result", "png." };
		for (String string : clutter) {
			if (filename.contains(string)) {
				filename = filename.replace(string, "");
			}
		}
		return filename;
	}

	/**
	 * Removes duplicated files from the map.
	 *
	 * @param fileMap
	 */
	private static void removeDuplicates(Map<String, File> fileMap) {
		Set<File> files = new HashSet<File>();
		Vector<String> removeKeys = new Vector<String>();
		for (Map.Entry<String, File> entry : fileMap.entrySet()) {
			if (files.contains(entry.getValue())) {
				removeKeys.add(entry.getKey());
			} else {
				files.add(entry.getValue());
			}
		}
		for (String key : removeKeys) {
			fileMap.remove(key);
		}
	}

	/**
	 * Return a Map with all TraceResults from property file
	 *
	 * @return
	 */
	public static Map<String, AbstractTraceResult> getTraceResults() {
		return getTraceObjectResultMap(getResults());
	}

	/**
	 * Return TraceObjectResult for given file (read it or return cached version
	 * if available)
	 *
	 * @param file
	 * @return
	 */
	public static AbstractTraceResult getTraceObjectResult(File file) {
		if (resultCache.containsKey(file)) {
			logger.info("Returning cached result: " + file);
			return resultCache.get(file);
		}
		try {
			Date date = new Date(file.lastModified());
			logger.info("Reading: " + file + " \t Modification Date: " + date);
			Object o = ObjectIO.getFromFile(file);
			if (!(o instanceof AbstractTraceResult)) {
				throw new RuntimeException(
						"Expected AbstractTraceResult, got: "
								+ o.getClass().getCanonicalName());
			}
			AbstractTraceResult traceObjectResult = (AbstractTraceResult) o;
			resultCache.put(file, traceObjectResult);
			return traceObjectResult;
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * Return all result files of a directory (files ending with .ser). HashMap
	 * consists of pairs (filename ( with removed clutter) ,file)
	 *
	 * @param dir
	 * @return
	 */
	private static Map<String, File> getFilesFromDirectory(File dir) {
		File[] files = dir.listFiles(new FilenameFilter() {
			public boolean accept(File directory, String name) {
				if (name.endsWith(".ser")) {
					return true;
				}
				return false;
			}
		});
		Map<String, File> resultMap = new HashMap<String, File>();
		for (File file : files) {
			resultMap.put(removeClutter(file.getName()), file);
		}
		return resultMap;
	}

}
