package org.softevo.apibirthmark.tracing.result.process.experiments;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Formatter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Class that stores the results and allows to print them in a table.
 *
 * @author schuler
 *
 */
public class ResultTable {

	private Map<String, Map<String, Float>> rowMap = new HashMap<String, Map<String, Float>>();

	private static final String ENTRY_NUMBER_FORMAT = "%.2f";

	private static final int NAME_LENGTH = 30;

	private static final String CONSOLE_CELL_FORMAT = "  %-" + NAME_LENGTH
			+ "s";

	private static final String LATEX_CELL_FORMAT = "  %-" + NAME_LENGTH + "s ";

	/**
	 * Enters a row in into the table
	 *
	 * @param string
	 *            key of the row
	 * @param map
	 *            Map that contains the results
	 */
	public void putRow(String string, Map<String, Float> map) {
		rowMap.put(string, map);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return toString(CONSOLE_CELL_FORMAT, "\n", "");
	}

	private int getRowNumber() {
		Iterator<Map<String, Float>> iterator = rowMap.values().iterator();
		if (iterator.hasNext()) {
			return iterator.next().size();
		}
		return 0;
	}

	/**
	 * Return a latex representation of this table
	 *
	 * @return A String that contains the latex representation of this table
	 */
	public String toLatexString() {
		StringBuilder buffer = new StringBuilder();
		buffer.append("\\begin{tabular}{l");
		for (int i = 0; i < getRowNumber(); i++) {
			buffer.append('r');
		}
		buffer.append("}\n");
		buffer.append(toString(LATEX_CELL_FORMAT, "\\\\ \n", "&"));
		buffer.append("\\end{tabular} \n");
		String latexString = buffer.toString();
		return latexString.replace('_', ' ');

	}

	/**
	 * Return a String that contains the contents of this table using the
	 * different format Strings.
	 *
	 * @param cellFormat
	 *            Format of a cell.
	 * @param newline
	 *            Newline character.
	 * @param cellDivider
	 *            String that is used to divide the cells.
	 * @return A formated String that contains the contents of this table.
	 */
	public String toString(String cellFormat, String newline, String cellDivider) {
		StringBuilder buffer = new StringBuilder();
		Formatter formatter = new Formatter(buffer);
		formatter.format(cellFormat, "");
		List<String> columnOrderList = getColumnOrderList();
		List<String> rowOrderList = getRowOrderList(columnOrderList);
		for (String columnOrder : columnOrderList) {
			buffer.append(cellDivider);
			formatter.format(cellFormat, columnOrder);
		}
		buffer.append(newline);
		for (String rowOrder : rowOrderList) {
			formatter.format(cellFormat, rowOrder);
			String rowKey = rowOrder;
			for (String columnOrder : columnOrderList) {
				buffer.append(cellDivider);
				formatter.format(cellFormat, String
						.format(ENTRY_NUMBER_FORMAT, Math.round(rowMap.get(
								rowKey).get(columnOrder) * 100) * 0.01f));

			}
			buffer.append(newline);
		}
		return buffer.toString();
	}

	private List<String> getRowOrderList(List<String> columnOrderVector) {
		List<String> rowOrderList = new ArrayList<String>();
		Set<String> rowSet = new HashSet<String>(rowMap.keySet());
		for (String columnOrder : columnOrderVector) {
			if (rowSet.contains(columnOrder)) {
				rowOrderList.add(columnOrder);
				rowSet.remove(columnOrder);
			}
		}
		for (String rowKey : rowSet) {
			rowOrderList.add(rowKey);
		}
		return rowOrderList;
	}

	private List<String> getColumnOrderList() {
		List<String> columnOrderList = new ArrayList<String>();
		String key = rowMap.keySet().toArray(new String[0])[0];
		for (Map.Entry<String, Float> entry : rowMap.get(key).entrySet()) {
			columnOrderList.add(entry.getKey());
		}
		Collections.sort(columnOrderList);
		return columnOrderList;
	}
}