package org.softevo.apibirthmark.transformation;

import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.jar.JarFile;

import org.apache.log4j.Logger;
import org.softevo.sibrelib.source.ClassFile;
import org.softevo.sibrelib.source.ClassFileSource;
import org.softevo.sibrelib.source.DirectorySource;
import org.softevo.sibrelib.source.JarSource;

import org.softevo.apibirthmark.tracing.mapping.NameToIDMap;
import org.softevo.apibirthmark.transformation.bytecodeTransformer.BytecodeTransformer;
import org.softevo.apibirthmark.transformation.bytecodeTransformer.HierarchyTransformer;
import org.softevo.apibirthmark.transformation.bytecodeTransformer.InsertCallsTransformer;
import org.softevo.apibirthmark.transformation.bytecodeTransformer.ObservedTransformer;
import org.softevo.apibirthmark.transformation.bytecodeTransformer.ProxyTransformer;
import org.softevo.apibirthmark.asmadapters.calls.observed.ExcludeCalls;
import org.softevo.apibirthmark.asmadapters.calls.observed.ObservedCalls;
import org.softevo.apibirthmark.asmadapters.calls.observed.ObservedCallsFromMap;
import org.softevo.apibirthmark.asmadapters.proxies.classHierarchy.ClassHierarchy;
import org.softevo.apibirthmark.constants.Constants;

/**
 * The FileTransformer class does instrumentation of files containing bytecode
 *
 * @author schuler
 *
 */
public class FileTransformer {

	private static Logger logger = Logger.getLogger(FileTransformer.class);

	private JarFile jar;

	private File directory;

	/**
	 * Creates new FileTransformer for a given jar file or a directory
	 * containing class files.
	 *
	 * @param file
	 *            directory or jar file
	 * @throws IOException
	 */
	public FileTransformer(File file) throws IOException {
		if (file.isDirectory()) {
			directory = file;
			jar = null;
		} else {
			directory = null;
			jar = new JarFile(file);
		}
	}

	/**
	 * Do the instrumentation and store the result at the given destination
	 *
	 * @param destination
	 *            place of the resulting jar file
	 * @param transformationType
	 *            type of transformation
	 * @see BytecodeTransformer
	 *
	 */

	public void transform(File destination,
			BytecodeTransformer bytecodeTransformer) {
		ClassFileSource classFileSource = getClassFileSource(destination);
		process(classFileSource, bytecodeTransformer);
		try {
			classFileSource.saveModifiedVersion();
		} catch (IOException e) {
			logger.error("Saving File " + destination + " failed");
			e.printStackTrace();
		}
	}

	/**
	 * Process all classes but do not store result.
	 *
	 * @param bytecodeTransformer
	 *            type of transformation
	 */
	public void process(BytecodeTransformer bytecodeTransformer) {
		ClassFileSource classFileSource = getClassFileSource(null);
		process(classFileSource, bytecodeTransformer);
	}

	/**
	 * Apply given BytecodeTransformer to all classes from given
	 * classFileSource.
	 *
	 * @param classFileSource
	 * @param bytecodeTransformer
	 */
	private void process(ClassFileSource classFileSource,
			BytecodeTransformer bytecodeTransformer) {
		Iterator it = classFileSource.getClassIterator();
		int classCount = 0;
		while (it.hasNext()) {
			ClassFile element = (ClassFile) it.next();
			byte[] transformedBytecode = bytecodeTransformer
					.transformBytecode(element.getClassBytes());
			element.setModifiedClassBytes(transformedBytecode);
			classCount++;
		}
		logger.info(classCount + " classes where processed");
	}

	/**
	 * Return a ClassFileSouce of either the a jar file or a directory that is
	 * initialized with the given destination
	 *
	 * @param destination
	 *            destination of the instrumented file(s)
	 * @return ClassFileSource for jar or directory
	 */
	private ClassFileSource getClassFileSource(File destination) {
		try {
			if (directory != null) {
				return new DirectorySource(directory, destination);
			} else {
				return new JarSource(jar, destination);
			}
		} catch (IOException e) {
			e.printStackTrace();
			throw new RuntimeException(e.getMessage());
		}

	}

	/*
	 * =========================================================================
	 *
	 *
	 * STATIC
	 *
	 *
	 * =========================================================================
	 */

	/**
	 * Transform an API (generate additional Proxy methods)
	 *
	 * @param source
	 *            API source
	 * @param destination
	 *            destination of modified file
	 * @param idMapDestination
	 *            destination for ID file
	 * @param maxKeyFile
	 *            idMap to get max method-ID
	 * @param idMapSource
	 *            source NameToIDMap if specified this NameToIDMap will be
	 *            extended otherwise a new NameToIDMap is created
	 */
	public static void transformAPI(String source, String destination,
			String idMapDestination, String maxKeyFile, String idMapSource) {
		NameToIDMap apiIDMap = null;
		if (idMapSource != null) {
			try {
				logger.info("Reading IDMap from" + idMapSource);
				apiIDMap = NameToIDMap.getFromFile(idMapSource);
			} catch (IOException e) {
				e.printStackTrace();
				throw new RuntimeException("Couldn't load Idmap from file:\n"
						+ idMapSource);
			}
		} else {
			apiIDMap = new NameToIDMap();
		}
		if (maxKeyFile != null) {
			try {
				int maxKey = 0;
				NameToIDMap maxKeyMap = NameToIDMap.getFromFile(maxKeyFile);
				maxKey = maxKeyMap.getMaxClassKey();
				apiIDMap.setMaxClassKey(maxKey);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		logger.info("Inserted Calls");
		BytecodeTransformer proxyTransformer = new ProxyTransformer(apiIDMap);
		transformBytecode(source, destination, proxyTransformer);
		if (idMapDestination != null) {
			try {
				apiIDMap.saveToFile(idMapDestination);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * Transform classes such that calls to the API are traced. Inserts calls to
	 * proxy methods.
	 *
	 * @param fileSource
	 *            source file
	 * @param fileDestination
	 *            destination file where file is saved
	 * @param apiIDMapSource
	 *            indicates which calls should be observed
	 * @param observedIDMapDestination
	 *            where to save own IDMap
	 * @param maxKeyFile
	 *            idMap to get max method-ID
	 */
	public static void transformObserved(String fileSource,
			String fileDestination, String apiIDMapSource,
			String observedIDMapDestination, String maxKeyFile) {
		NameToIDMap apiIDMap = null;
		try {
			apiIDMap = NameToIDMap.getFromFile(apiIDMapSource);
		} catch (IOException e) {
			e.printStackTrace();
		}
		NameToIDMap callerIDMap = new NameToIDMap();
		int maxKey = apiIDMap.getMaxClassKey();
		if (maxKeyFile != null) {
			try {
				NameToIDMap maxKeyMap = NameToIDMap.getFromFile(maxKeyFile);
				maxKey = maxKeyMap.getMaxClassKey();
			} catch (IOException e) {
				e.printStackTrace();
			}

		}
		callerIDMap.setMaxClassKey(maxKey);
		transformObservedSpecifyCalls(fileSource, fileDestination, apiIDMap,
				callerIDMap);
		if (observedIDMapDestination != null) {
			try {
				callerIDMap.saveToFile(observedIDMapDestination);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * Transforms a library that should be used but calls should not be traced
	 * @param source
	 * @param destination
	 * @param apiIDMap
	 * @param idMapDestination
	 * @param maxKeyFile
	 */
	public static void transformIntermediateLibrary(String source,
			String destination, String apiIDMap, String idMapDestination,
			String maxKeyFile) {
		NameToIDMap nameToIDMap = new NameToIDMap();
		NameToIDMap apiNameToIDMap = null;
		try {
			apiNameToIDMap = NameToIDMap.getFromFile(apiIDMap);
		} catch (IOException e1) {
			logger.error("Couldn't load api ID map");
			e1.printStackTrace();
		}
		if (maxKeyFile != null) {
			try {
				int maxKey = 0;
				NameToIDMap maxKeyMap = NameToIDMap.getFromFile(maxKeyFile);
				maxKey = maxKeyMap.getMaxClassKey();
				nameToIDMap.setMaxClassKey(maxKey);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		ClassHierarchy classHierarchy = new ClassHierarchy();
		HierarchyTransformer hierarchyTransformer = new HierarchyTransformer(
				classHierarchy);
		transformBytecode(source, destination, hierarchyTransformer);
		ObservedTransformer observedBytecodeTransformer = new ObservedTransformer(
				nameToIDMap, apiNameToIDMap, classHierarchy);
		transformBytecode(source, destination, observedBytecodeTransformer);
		if (idMapDestination != null) {
			try {
				nameToIDMap.saveToFile(idMapDestination);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}

	}

	private static void processBytecode(String origin,
			BytecodeTransformer bytecodeTransformer) {
		File originFile = new File(origin);
		if (originFile.exists()) {
			try {
				FileTransformer jarTransformer = new FileTransformer(originFile);
				jarTransformer.process(bytecodeTransformer);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	private static void transformBytecode(String origin, String destination,
			BytecodeTransformer bytecodeTransformer) {
		transformBytecode(origin, destination, bytecodeTransformer, true);
	}

	private static void transformBytecode(String origin, String destination,
			BytecodeTransformer bytecodeTransformer, boolean checkIfWritten) {
		File originFile = new File(origin);
		if (originFile.exists()) {
			File destinationFile = new File(destination);
			FileTransformer jarTransformer;
			try {
				jarTransformer = new FileTransformer(originFile);
				jarTransformer.transform(destinationFile, bytecodeTransformer);
			} catch (IOException e) {
				logger.error("Instrumentation failed");
				e.printStackTrace();
			}
			if (checkIfWritten) {
				checkIfWritten(destinationFile);
			}
		} else {
			logger.error("File does not exist:\n" + origin);
		}
	}

	private static void checkIfWritten(File destinationFile) {
		if (destinationFile.exists()) {
			if (destinationFile.isFile()) {
				logger.info("File " + destinationFile.toString() + " written.");
			}
			if (destinationFile.isDirectory()) {
				logger.info("Directory " + destinationFile.toString()
						+ " written.");
			}
		}
	}

	/**
	 * Transform classes such that calls to the API are traces. Inserts calls to
	 * proxy methods. If calls should be excluded special instrumentation is
	 * used
	 *
	 *
	 * @param source
	 * @param dest
	 * @param apiIDMap
	 * @param callerIDMap
	 */
	private static void transformObservedSpecifyCalls(String source,
			String dest, NameToIDMap apiIDMap, NameToIDMap callerIDMap) {
		if (System.getProperty(Constants.KEY_OBSERVED_CALLS) != null) {
			if (System.getProperty(Constants.KEY_OBSERVED_CALLS).equals(
					"exclude")) {
				logger.info("Setting Exclude Handler");
				observedTransformation(source, dest, callerIDMap, apiIDMap,
						new ExcludeCalls(apiIDMap));
			} else {
				throw new RuntimeException("value:"
						+ System.getProperty(Constants.KEY_OBSERVED_CALLS)
						+ " not known for key " + Constants.KEY_OBSERVED_CALLS);
			}
		} else {
			logger.info("No call handler specified");
			observedTransformation(source, dest, callerIDMap, apiIDMap,
					new ObservedCallsFromMap(apiIDMap));
		}
	}

	/**
	 * @param source Source: classes to instrument
	 * @param dest Destination of the instrumented classes
	 * @param callerIDMap Map to insert class and method IDs
	 * @param apiIDMap  Map that contains class and method IDs of the API
	 * @param observedCalls An Adapter that decides which calls should be observed
	 */
	private static void observedTransformation(String source, String dest,
			NameToIDMap callerIDMap, NameToIDMap apiIDMap,
			ObservedCalls observedCalls) {
		String temp = Constants.OUTPUT_DIRECTORY + "TEMP.jar";
		if ((new File(source)).isDirectory()) {
			temp = Constants.OUTPUT_DIRECTORY + "classes/";
		}
		ClassHierarchy classHierarchy = new ClassHierarchy();
		BytecodeTransformer hierarchyTransformer = new HierarchyTransformer(
				classHierarchy);
		transformBytecode(source, temp, new InsertCallsTransformer(callerIDMap,
				observedCalls));
		processBytecode(temp, hierarchyTransformer);
		transformBytecode(temp, dest, new ObservedTransformer(callerIDMap,
				apiIDMap, classHierarchy));
		File tempFile = new File(temp);
		if (!tempFile.isDirectory()) {
			tempFile.delete();
		}
	}
}