package org.softevo.apibirthmark.validation;

/**
 * Class represents the API call graph that models program API interaction. From
 * this class an expected trace result can be generated and Java code that
 * produces the program API interaction described by this graph.
 *
 * @author schuler
 *
 */
public class CallGraph {
	/**
	 * Represents an API class
	 *
	 * @author schuler
	 *
	 */
	static class APIClassNode {
		public String name;

		public APIMethodNode[] methods;

		public int id;

		public int constructorID;

		public APIClassNode(String name, APIMethodNode[] methods, int id) {
			this.name = name;
			this.methods = methods;
			this.id = id;
		}

	}

	/**
	 * Represents an API method contained in a class
	 *
	 * @author schuler
	 *
	 */
	static class APIMethodNode {
		public String name;

		public int id;

		public APIMethodNode(String name, int id) {
			this.name = name;
			this.id = id;
		}

	}

	/**
	 * Represents a class of a program that produces API interaction
	 *
	 * @author schuler
	 *
	 */
	static class CallingClassNode {
		public String name;

		public int id;

		public CallingMethodNode[] methods;

		public CallingClassNode(String name, CallingMethodNode[] methods, int id) {
			this.name = name;
			this.methods = methods;
			this.id = id;
		}
	}

	/**
	 * Represents a method of a program that produces API interaction
	 *
	 * @author schuler
	 *
	 */
	static class CallingMethodNode {
		public String name;

		public int id;

		public CallingObjectNode[] callingObjects;

		public CallingMethodNode(String name,
				CallingObjectNode[] callingObjects, int id) {
			this.name = name;
			this.id = id;
			this.callingObjects = callingObjects;
		}

	}

	/**
	 * Represents an object of an API class that exists in the program/
	 *
	 * @author schuler
	 *
	 */
	static class CallingObjectNode {
		public String name;

		public String type;

		public APIMethodNode[] methods;

		public APIClassNode apiClassNode;

		public CallingObjectNode(APIClassNode apiClassNode, String name,
				APIMethodNode[] callMethods) {
			this.apiClassNode = apiClassNode;
			this.type = apiClassNode.name;
			this.name = name;
			this.methods = callMethods;
		}
	}

	/**
	 * Represents the main class of the program. That is executed to produce the
	 * program API interaction of this graph.
	 *
	 * @author schuler
	 *
	 */
	static class MainCallingClass {
		public String name;

		public MainObjectNode[] objects;

		public MainCallingClass(String name, MainObjectNode[] mainObjectNodes) {
			this.name = name;
			this.objects = mainObjectNodes;
		}
	}

	/**
	 * Represents an object in the main class which is an instance of a program
	 * class and on which program methods are called.
	 *
	 * @author schuler
	 *
	 */
	static class MainObjectNode {
		public String name;

		public String type;

		public CallingMethodNode[] callingMethodNodes;

		public CallingClassNode callingClassNode;

		public MainObjectNode(String name, CallingClassNode callingClassNode,
				CallingMethodNode[] callingMethodNodes) {
			this.name = name;
			this.callingClassNode = callingClassNode;
			this.type = callingClassNode.name;
			this.callingMethodNodes = callingMethodNodes;
		}

	}
}
