package org.softevo.apibirthmark.validation;

import java.io.FileWriter;
import java.io.IOException;
import java.util.Iterator;
import java.util.Vector;

import org.softevo.apibirthmark.tracing.Tracer;
import org.softevo.apibirthmark.tracing.Tracer.TraceObject;
import org.softevo.apibirthmark.tracing.mapping.IDToNameMap;
import org.softevo.apibirthmark.tracing.mapping.NameToIDMap;
import org.softevo.apibirthmark.util.ObjectIO;
import org.softevo.apibirthmark.constants.Constants;

/**
 * Class compares the expected trace with the trace derived from translating the
 * API call graph into a Java program and tracing its program API interaction.
 *
 * @author schuler
 *
 */
public class CompareTraces {

	private IDToNameMap expectedIDMap;

	private IDToNameMap observedDMap;

	private boolean verbose = false;

	/**
	 * Compare the expected and observed trace. If there are differences a
	 * RuntimeException is thrown
	 *
	 * @throws IOException Thrown if the trace can not be read
	 * @throws ClassNotFoundException if there are problems deserializing the traces
	 */
	public void compareTraces() throws IOException, ClassNotFoundException {
		Vector expectedVector = (Vector) ObjectIO
				.getFromFile(Constants.VALIDATION_EXPECTED_TRACE_FILE);
		Vector observedVector = (Vector) ObjectIO
				.getFromFile(Constants.VALIDATION_OBSERVED_TRACE_FILE);
		expectedIDMap = NameToIDMap
				.getIDToNameMapfromNameToIDFile(Constants.VALIDATION_EXPECTED_IDMAP);
		observedDMap = NameToIDMap
				.getIDToNameMapfromNameToIDFile(Constants.VALIDATION_OBSERVED_IDMAP);
		if (observedVector.size() != expectedVector.size()) {
			throw new RuntimeException(
					"observedVector.size() != expectedVector.size()");
		}
		Iterator expectedIter = expectedVector.iterator();
		int callsEqual = 0;
		System.out.println("Start Comparing");
		FileWriter observedFW = new FileWriter(
				Constants.VALIDATION_OUTPUT_DIRECTORY + "observed.txt");
		FileWriter expectedFW = new FileWriter(
				Constants.VALIDATION_OUTPUT_DIRECTORY + "expected.txt");
		for (Iterator iter = observedVector.iterator(); iter.hasNext();) {
			Tracer.TraceObject observedElement = (Tracer.TraceObject) iter
					.next();
			Tracer.TraceObject expectedElement = (Tracer.TraceObject) expectedIter
					.next();
			observedFW.write(getTraceString(observedElement, observedDMap));
			expectedFW.write(getTraceString(expectedElement, expectedIDMap));
			if (areEqual(observedElement, expectedElement)) {
				callsEqual++;
			} else {
				throw new RuntimeException("Traced elements not equal \nExpected:\n"
						+ expectedElement + "\nObserved\n" + observedElement);
			}
		}
		observedFW.close();
		expectedFW.close();
		System.out.println("Finished\n" + callsEqual + " equal trace entries");
	}

	private String getTraceString(TraceObject traceObject, IDToNameMap idMap) {
		StringBuilder buffer = new StringBuilder();
		String spacer = " - ";
		buffer.append(idMap.getClassnameByID(traceObject.calleeClassID))
				.append(spacer);
		buffer.append(idMap.getMethodnameByID(traceObject.calleeMethodID))
				.append(spacer);
		buffer.append(idMap.getClassnameByID(traceObject.callerClassID))
				.append(spacer);
		buffer.append(idMap.getMethodnameByID(traceObject.callerMethodID))
				.append(spacer);
		return buffer.append('\n').toString();
	}

	private boolean areEqual(TraceObject observedElement,
			TraceObject expectedElement) {
		if (!observedDMap
				.getClassnameByID(observedElement.calleeClassID)
				.equals(
						expectedIDMap
								.getClassnameByID(expectedElement.calleeClassID))) {
			return false;
		}
		if (!observedDMap
				.getClassnameByID(observedElement.callerClassID)
				.equals(
						expectedIDMap
								.getClassnameByID(expectedElement.callerClassID))) {
			return false;
		}
		if (verbose) {
			System.out.println("ObservedMethod: ("
					+ observedElement.calleeMethodID
					+ ") "
					+ observedDMap.getMethodnameByID(
							observedElement.calleeClassID,
							observedElement.calleeMethodID));
			System.out.println("ExpectedMethod: ("
					+ expectedElement.calleeMethodID
					+ ") "
					+ expectedIDMap.getMethodnameByID(
							expectedElement.calleeClassID,
							expectedElement.calleeMethodID));
		}

		if (!observedDMap.getMethodnameByID(observedElement.calleeClassID,
				observedElement.calleeMethodID).equals(
				expectedIDMap.getMethodnameByID(expectedElement.calleeClassID,
						expectedElement.calleeMethodID))) {
			return false;
		}
		if (!observedDMap.getMethodnameByID(observedElement.callerClassID,
				observedElement.callerMethodID).equals(
				expectedIDMap.getMethodnameByID(expectedElement.callerClassID,
						expectedElement.callerMethodID))) {
			return false;
		}
		return true;
	}

	/**
	 * Used from CL to compare the expected and observed trace.
	 *
	 * @param args
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	public static void main(String[] args) throws IOException,
			ClassNotFoundException {
		CompareTraces cp = new CompareTraces();
		cp.compareTraces();
	}
}
