package org.softevo.apibirthmark.validation;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Random;

import org.antlr.stringtemplate.StringTemplate;
import org.antlr.stringtemplate.StringTemplateGroup;
import org.antlr.stringtemplate.language.AngleBracketTemplateLexer;
import org.apache.log4j.Logger;

import org.softevo.apibirthmark.tracing.mapping.NameToIDMap;
import org.softevo.apibirthmark.validation.CallGraph.APIClassNode;
import org.softevo.apibirthmark.validation.CallGraph.APIMethodNode;
import org.softevo.apibirthmark.validation.CallGraph.CallingClassNode;
import org.softevo.apibirthmark.validation.CallGraph.CallingMethodNode;
import org.softevo.apibirthmark.validation.CallGraph.CallingObjectNode;
import org.softevo.apibirthmark.validation.CallGraph.MainCallingClass;
import org.softevo.apibirthmark.validation.CallGraph.MainObjectNode;
import org.softevo.apibirthmark.constants.Constants;

/**
 * Class that generated the API call graph and produces the corresponding java
 * classes for it.
 *
 * @author schuler
 *
 */
public class ValidationClassGenerator {

	private static final String TEMPLATE_FILE = "template/validationClasses.stg";

	private static final String API_PACKAGE_NAME = "validationapi";

	private static final String API_CLASS_NAME_PREFIX = "APIClass";

	private static final String API_METHOD_NAME_PREFIX = "apiMethod";

	private static final String CALLING_OBJECT_NAME = "callingObject";

	private static final String CALLING_METHOD_NAME = "callingMethod";

	private static final String CALLING_CLASS_NAME = "CallingClass";

	private static final String MAIN_OBJECT_NAME = "mainObject";

	private static final String MAIN_CLASS_NAME = "MainClass";

	private static Logger logger = Logger
			.getLogger(ValidationClassGenerator.class);

	static Random random = new Random();

	private static APIMethodNode generateAPIMethodNode(String name, int id) {
		return new APIMethodNode(name, id);

	}

	private static APIMethodNode[] generateAPIMethodNodes(int numberOfMethods) {
		APIMethodNode[] apiMethodNodes = new APIMethodNode[numberOfMethods];
		for (int i = 0; i < apiMethodNodes.length; i++) {
			apiMethodNodes[i] = generateAPIMethodNode(API_METHOD_NAME_PREFIX
					+ getFormatedNumber(i), i);
		}
		return apiMethodNodes;
	}

	private static APIClassNode generateAPIClassNode(int id, int numberOfMethods) {
		APIMethodNode[] methods = generateAPIMethodNodes(numberOfMethods);
		return new APIClassNode(API_CLASS_NAME_PREFIX + getFormatedNumber(id),
				methods, id);
	}

	public static APIClassNode[] generateAPIClassNodes(int numberOfClasses,
			int numberOfMethods) {
		APIClassNode[] apiClasses = new APIClassNode[numberOfClasses];
		for (int i = 0; i < apiClasses.length; i++) {
			apiClasses[i] = generateAPIClassNode(i, numberOfMethods);
		}
		return apiClasses;
	}

	private static void updateAPIClassIDs(APIClassNode[] apiClasses,
			NameToIDMap idMap) {
		for (int i = 0; i < apiClasses.length; i++) {
			String apiClassName = API_PACKAGE_NAME + "/" + apiClasses[i].name;
			idMap.insertClass(apiClassName);
			apiClasses[i].id = idMap.getClassID(apiClassName);
			idMap.insertMethod(apiClassName, "<init>", "()V");
			apiClasses[i].constructorID = idMap.getClassMethodID(apiClassName,
					"<init>", "()V").getMethodID();
			for (int j = 0; j < apiClasses[i].methods.length; j++) {
				APIMethodNode method = apiClasses[i].methods[j];
				idMap.insertMethod(apiClassName, method.name, "(III)V");
				method.id = idMap.getClassMethodID(apiClassName, method.name,
						"(III)V").getMethodID();
			}
		}
	}

	private static CallingObjectNode generateCallingObjectNode(String name,
			APIClassNode apiClassNode, int numberOfCalls) {
		APIMethodNode[] callMethods = new APIMethodNode[numberOfCalls];
		for (int i = 0; i < callMethods.length; i++) {
			callMethods[i] = apiClassNode.methods[random
					.nextInt(apiClassNode.methods.length)];
		}
		return new CallingObjectNode(apiClassNode, name, callMethods);

	}

	private static CallingObjectNode[] generateCallingObjectNodes(
			APIClassNode[] apiClassNodes, int numberOfObjects, int numberOfCalls) {
		CallingObjectNode[] callingObjectNodes = new CallingObjectNode[numberOfObjects];
		for (int i = 0; i < callingObjectNodes.length; i++) {
			callingObjectNodes[i] = generateCallingObjectNode(
					CALLING_OBJECT_NAME + getFormatedNumber(i),
					apiClassNodes[random.nextInt(apiClassNodes.length)],
					numberOfCalls);
		}
		return callingObjectNodes;
	}

	private static CallingMethodNode generateCallingMethodNode(String name,
			int id, int numberOfObjects, int numberOfCalls,
			APIClassNode[] apiClassNodes) {
		return new CallingMethodNode(name, generateCallingObjectNodes(
				apiClassNodes, numberOfObjects, numberOfCalls), id);
	}

	private static CallingMethodNode[] generateCallingMethodNodes(
			int numberOfMethods, int numberOfObjects, int numberOfCalls,
			APIClassNode[] apiClassNodes) {
		CallingMethodNode[] callingMethodNodes = new CallingMethodNode[numberOfMethods];
		for (int i = 0; i < callingMethodNodes.length; i++) {
			callingMethodNodes[i] = generateCallingMethodNode(
					CALLING_METHOD_NAME + getFormatedNumber(i), i,
					numberOfObjects, numberOfCalls, apiClassNodes);
		}
		return callingMethodNodes;
	}

	private static CallingClassNode generateCallingClass(String name, int id,
			int numberOfMethods, int numberOfObjects, int numberOfCalls,
			APIClassNode[] apiClassNodes) {
		return new CallingClassNode(name,
				generateCallingMethodNodes(numberOfMethods, numberOfObjects,
						numberOfCalls, apiClassNodes), id);
	}

	private static CallingClassNode[] generateCallingClasses(
			int numberOfClasses, int numberOfMethods, int numberOfObjects,
			int numberOfCalls, APIClassNode[] apiClassNodes) {
		CallingClassNode[] callingClasses = new CallingClassNode[numberOfClasses];
		for (int i = 0; i < callingClasses.length; i++) {
			callingClasses[i] = generateCallingClass(CALLING_CLASS_NAME
					+ getFormatedNumber(i), i, numberOfMethods,
					numberOfObjects, numberOfCalls, apiClassNodes);
		}

		return callingClasses;
	}

	private static void updateCallingClassIDs(
			CallingClassNode[] callingClasses, NameToIDMap idMap) {

		for (int i = 0; i < callingClasses.length; i++) {
			String className = callingClasses[i].name;
			idMap.insertClass(className);
			callingClasses[i].id = idMap.getClassID(className);
			for (int j = 0; j < callingClasses[i].methods.length; j++) {
				CallingMethodNode method = callingClasses[i].methods[j];
				idMap.insertMethod(callingClasses[i].name, method.name, "()V");
				method.id = idMap.getClassMethodID(callingClasses[i].name,
						method.name, "()V").getMethodID();
			}
		}

	}

	private static MainObjectNode generateMainObjectNode(String name,
			int numberOfCalls, CallingClassNode callingClassNode) {
		CallingMethodNode[] calls = new CallingMethodNode[numberOfCalls];
		for (int i = 0; i < calls.length; i++) {
			calls[i] = callingClassNode.methods[random
					.nextInt(callingClassNode.methods.length)];
		}
		return new MainObjectNode(name, callingClassNode, calls);

	}

	private static MainCallingClass generateMainCallingClass(
			int numberOfObjects, int numberOfCalls,
			CallingClassNode[] callingClassNodes) {
		MainObjectNode[] mainObjectNodes = new MainObjectNode[numberOfObjects];
		for (int i = 0; i < mainObjectNodes.length; i++) {
			mainObjectNodes[i] = generateMainObjectNode(MAIN_OBJECT_NAME
					+ getFormatedNumber(i), numberOfCalls,
					callingClassNodes[random.nextInt(callingClassNodes.length)]);
		}

		return new MainCallingClass(MAIN_CLASS_NAME, mainObjectNodes);
	}

	/**
	 * Generates the API call graph, translates it into.
	 *
	 * @param args
	 * @throws IOException
	 * @throws URISyntaxException
	 */
	public static void main(String[] args) throws IOException,
			URISyntaxException {
		generateValidationFiles();
	}

	private static void generateValidationFiles() throws IOException, URISyntaxException, FileNotFoundException {
		APIClassNode[] apiClasses = generateAPIClassNodes(10, 10);
		NameToIDMap validationMap = new NameToIDMap();
		updateAPIClassIDs(apiClasses, validationMap);
		CallingClassNode[] callingClasses = generateCallingClasses(10, 10, 10,
				5, apiClasses);
		updateCallingClassIDs(callingClasses, validationMap);
		MainCallingClass mainCallingClass = generateMainCallingClass(20, 10,
				callingClasses);
		generateExpectedTrace(mainCallingClass);
		validationMap.saveToFile(Constants.VALIDATION_EXPECTED_IDMAP);
		StringTemplateGroup group = getTemplateGroup();
		writeAPIFiles(apiClasses, group);
		writeCallingClasses(callingClasses, mainCallingClass, group);
	}

	private static void generateExpectedTrace(MainCallingClass mainCallingClass) {
		TraceGenerator expectedTrace = new TraceGenerator();
		expectedTrace.generateExpectedTrace(mainCallingClass);
		expectedTrace.saveToFile(Constants.VALIDATION_EXPECTED_TRACE_FILE);
	}

	private static StringTemplateGroup getTemplateGroup() throws URISyntaxException, FileNotFoundException {
		URL url = ValidationClassGenerator.class.getResource(TEMPLATE_FILE);
		if (url == null) {
			logger.warn("could not find template file" + TEMPLATE_FILE);
		}
		File templateFile = new File(url.toURI());
		StringTemplateGroup group = new StringTemplateGroup(new FileReader(
				templateFile), AngleBracketTemplateLexer.class);
		return group;
	}

	private static void writeCallingClasses(CallingClassNode[] callingClasses, MainCallingClass mainCallingClass, StringTemplateGroup group) throws IOException {
		StringTemplate callingClassST = group.getInstanceOf("callingClass");
		callingClassST.setAttribute("apiPackageName", API_PACKAGE_NAME);

		StringTemplate mainCallingClassST = group
				.getInstanceOf("mainCallingClass");
		mainCallingClassST.setAttribute("mainClass", mainCallingClass);


		for (int i = 0; i < callingClasses.length; i++) {
			callingClassST.setAttribute("class", callingClasses[i]);
			writeJavaFile(callingClasses[i].name, callingClassST.toString());
			callingClassST.removeAttribute("class");
		}
		writeJavaFile(MAIN_CLASS_NAME, mainCallingClassST.toString());
	}

	private static void writeAPIFiles(APIClassNode[] apiClasses,
			StringTemplateGroup group ) throws IOException {
		StringTemplate apiClassST = group.getInstanceOf("class");
		apiClassST.setAttribute("packageName", API_PACKAGE_NAME);
		String apiPath = API_PACKAGE_NAME + "/";
		File path = new File(Constants.VALIDATION_OUTPUT_DIRECTORY + apiPath);
		if (!path.exists()) {
			path.mkdir();
		}
		for (int i = 0; i < apiClasses.length; i++) {
			apiClassST.setAttribute("class", apiClasses[i]);
			writeJavaFile(apiPath + apiClasses[i].name, apiClassST.toString());
			apiClassST.removeAttribute("class");
		}
	}

	private static String getFormatedNumber(int number) {
		return String.format("%03d", number);
	}

	private static void writeJavaFile(String name, String content)
			throws IOException {
		File sourceFile = new File(Constants.VALIDATION_OUTPUT_DIRECTORY + name
				+ ".java");
		if (!sourceFile.exists()) {
			sourceFile.createNewFile();
		}
		FileWriter classWriter = new FileWriter(sourceFile);
		System.out.println("Writing " + sourceFile);
		classWriter.write(content);
		classWriter.close();
	}

}
